"""Instrument driver for the Thorlabs K10CR1/M motorized rotational mount.

This driver communicates with the device via a USB serial port,
using the Thorlabs APT protocol. For details, see the document
"Thorlabs APT Controllers Host-Controller Communications Protocol",
issue 25 from Thorlabs.

This driver has only been tested under Linux. In principle it should also
work under Windows, but that would require somehow creating a virtual COM port
for the internal USB serial port in the instrument.
"""

import ctypes
import contextlib
import enum
import logging
import time
from typing import Any, ClassVar, NamedTuple, Type

from qmi.core.context import QMI_Context
from qmi.core.exceptions import QMI_InstrumentException, QMI_TimeoutException
from qmi.core.instrument import QMI_Instrument, QMI_InstrumentIdentification
from qmi.core.rpc import rpc_method
from qmi.core.transport import create_transport

# Global variable holding the logger for this module.
_logger = logging.getLogger(__name__)


_APT_HOST_ADDRESS = 0x01    # Address of the computer in the APT protocol.
_APT_DEVICE_ADDRESS = 0x50  # Address of the device in the APT protocol.


class _AptMessage(ctypes.LittleEndianStructure):
    """Base class for APT protocol messages."""

    _pack_ = 1

    @classmethod
    def create(cls, **kwargs: Any) -> "_AptMessage":
        """Return a new message instance.

        Message ID, destination address, source address and data length
        (if applicable) will already be filled in. Other fields may be
        specified as keyword arguments.
        """
        message_size = ctypes.sizeof(cls)
        if message_size > 6:
            # This is a long APT message (header + data).
            # Long APT messages are identified by bit 7 in the destination field.
            return cls(
                message_id=cls.MESSAGE_ID,
                data_length=(message_size - 6),
                dest=(_APT_DEVICE_ADDRESS | 0x80),
                source=_APT_HOST_ADDRESS,
                **kwargs
            )
        else:
            # This is a short APT message (header only).
            return cls(
                message_id=cls.MESSAGE_ID,
                dest=_APT_DEVICE_ADDRESS,
                source=_APT_HOST_ADDRESS,
                **kwargs
            )


class _AptMessageHeader(_AptMessage):
    """Generic message header for the APT protocol.

    This is also the base class for long messages (header + data).
    """
    _pack_ = 1
    _fields_: ClassVar[list[tuple[str, type]]] = [
        ('message_id',  ctypes.c_uint16),
        ('data_length', ctypes.c_uint16),
        ('dest',        ctypes.c_uint8),
        ('source',      ctypes.c_uint8)
    ]


def _apt_short_message_fields(fields: list[tuple[str, type]]) -> list[tuple[str, type]]:
    """Helper function to create the field list for a short APT message.

    The message consists of only a 6-byte header, including 2 bytes space
    for optional parameter fields.
    """

    # APT short message format:
    #   2 bytes:  uint16  message_id
    #   2 bytes:  uint16  data_length
    #   1 byte:           param1
    #   1 byte:           param2
    #   1 byte:   uint8   dest
    #   1 byte:   uint8   source

    # Calculate size of data fields.
    data_length = 0
    for (par_name, par_type) in fields:
        data_length += ctypes.sizeof(par_type)
    assert data_length <= 2

    # Build field list for ctypes structure.
    all_fields: list[tuple[str, type]] = []
    all_fields.append(("message_id", ctypes.c_uint16))
    all_fields += fields
    if data_length < 2:
        # Add dummy field if the message uses fewer than 2 bytes for parameters.
        all_fields.append(("_dummy", (2 - data_length) * ctypes.c_uint8))
    all_fields.append(("dest", ctypes.c_uint8))
    all_fields.append(("source", ctypes.c_uint8))

    return all_fields


# APT messages for the K10CR1 device.
# See the "Thorlabs APT Controllers Host-Controller Communications Protocol".

class _AptMsgIdentify(_AptMessage):
    MESSAGE_ID = 0x0223
    _fields_ = _apt_short_message_fields([("chan_ident", ctypes.c_uint8)])

class _AptMsgSetChanEnableState(_AptMessage):
    MESSAGE_ID = 0x0210
    _fields_ = _apt_short_message_fields([("chan_ident", ctypes.c_uint8),
                                          ("enable_state", ctypes.c_uint8)])

class _AptMsgReqChanEnableState(_AptMessage):
    MESSAGE_ID = 0x0211
    _fields_ = _apt_short_message_fields([("chan_ident", ctypes.c_uint8)])

class _AptMsgGetChanEnableState(_AptMessage):
    MESSAGE_ID = 0x0212
    _fields_ = _apt_short_message_fields([("chan_ident", ctypes.c_uint8),
                                          ("enable_state", ctypes.c_uint8)])

class _AptMsgHwReqInfo(_AptMessage):
    MESSAGE_ID = 0x0005
    _fields_ = _apt_short_message_fields([])

class _AptMsgHwGetInfo(_AptMessageHeader):
    MESSAGE_ID = 0x0006
    _fields_ = [("serial_number",   ctypes.c_uint32),
                ("model_number",    8 * ctypes.c_char),
                ("type_",           ctypes.c_uint16),
                ("fw_version",      4 * ctypes.c_uint8),
                ("internal_use",    60 * ctypes.c_uint8),
                ("hw_version",      ctypes.c_uint16),
                ("mod_state",       ctypes.c_uint16),
                ("nchs",            ctypes.c_uint16)]

class _AptMsgReqPosCounter(_AptMessage):
    MESSAGE_ID = 0x0411
    _fields_ = _apt_short_message_fields([("chan_ident", ctypes.c_uint8)])

class _AptMsgGetPosCounter(_AptMessageHeader):
    MESSAGE_ID = 0x0412
    _fields_ = [("chan_ident",      ctypes.c_uint16),
                ("position",        ctypes.c_int32)]

class _AptMsgSetVelParams(_AptMessageHeader):
    MESSAGE_ID = 0x0413
    _fields_ = [("chan_ident",      ctypes.c_uint16),
                ("min_vel",         ctypes.c_int32),
                ("accel",           ctypes.c_int32),
                ("max_vel",         ctypes.c_int32)]

class _AptMsgReqVelParams(_AptMessage):
    MESSAGE_ID = 0x0414
    _fields_ = _apt_short_message_fields([("chan_ident", ctypes.c_uint8)])

class _AptMsgGetVelParams(_AptMessageHeader):
    MESSAGE_ID = 0x0415
    _fields_ = [("chan_ident",      ctypes.c_uint16),
                ("min_vel",         ctypes.c_int32),
                ("accel",           ctypes.c_int32),
                ("max_vel",         ctypes.c_int32)]

class _AptMsgSetGenMoveParams(_AptMessageHeader):
    MESSAGE_ID = 0x043A
    _fields_ = [("chan_ident",      ctypes.c_uint16),
                ("backlash_dist",   ctypes.c_int32)]

class _AptMsgReqGenMoveParams(_AptMessage):
    MESSAGE_ID = 0x043B
    _fields_ = _apt_short_message_fields([("chan_ident", ctypes.c_uint8)])

class _AptMsgGetGenMoveParams(_AptMessageHeader):
    MESSAGE_ID = 0x043C
    _fields_ = [("chan_ident",      ctypes.c_uint16),
                ("backlash_dist",   ctypes.c_int32)]

class _AptMsgSetHomeParams(_AptMessageHeader):
    MESSAGE_ID = 0x0440
    _fields_ = [("chan_ident",      ctypes.c_uint16),
                ("home_dir",        ctypes.c_uint16),
                ("limit_switch",    ctypes.c_uint16),
                ("home_velocity",   ctypes.c_int32),
                ("offset_dist",     ctypes.c_int32)]

class _AptMsgReqHomeParams(_AptMessage):
    MESSAGE_ID = 0x0441
    _fields_ = _apt_short_message_fields([("chan_ident", ctypes.c_uint8)])

class _AptMsgGetHomeParams(_AptMessageHeader):
    MESSAGE_ID = 0x0442
    _fields_ = [("chan_ident",      ctypes.c_uint16),
                ("home_dir",        ctypes.c_uint16),
                ("limit_switch",    ctypes.c_uint16),
                ("home_velocity",   ctypes.c_int32),
                ("offset_dist",     ctypes.c_int32)]

class _AptMsgMoveHome(_AptMessage):
    MESSAGE_ID = 0x0443
    _fields_ = _apt_short_message_fields([("chan_ident", ctypes.c_uint8)])

class _AptMsgMoveHomed(_AptMessage):
    MESSAGE_ID = 0x0444
    _fields_ = _apt_short_message_fields([("chan_ident", ctypes.c_uint8)])

class _AptMsgMoveRelative(_AptMessageHeader):
    MESSAGE_ID = 0x0448
    _fields_ = [("chan_ident",      ctypes.c_uint16),
                ("rel_dist",        ctypes.c_int32)]

class _AptMsgMoveAbsolute(_AptMessageHeader):
    MESSAGE_ID = 0x0453
    _fields_ = [("chan_ident",      ctypes.c_uint16),
                ("abs_position",    ctypes.c_int32)]

class _AptMsgMoveCompleted(_AptMessageHeader):
    MESSAGE_ID = 0x0464
    _fields_ = [("chan_ident",      ctypes.c_uint16),
                ("position",        ctypes.c_int32),
                ("velocity",        ctypes.c_uint16),
                ("reserved",        ctypes.c_uint16),
                ("status_bits",     ctypes.c_uint32)]

class _AptMsgMoveStop(_AptMessage):
    MESSAGE_ID = 0x0465
    _fields_ = _apt_short_message_fields([("chan_ident", ctypes.c_uint8),
                                          ("stop_mode", ctypes.c_uint8)])

class _AptMsgMoveStopped(_AptMessageHeader):
    MESSAGE_ID = 0x0466
    _fields_ = [("chan_ident",      ctypes.c_uint16),
                ("position",        ctypes.c_int32),
                ("velocity",        ctypes.c_uint16),
                ("reserved",        ctypes.c_uint16),
                ("status_bits",     ctypes.c_uint32)]

class _AptMsgReqStatusBits(_AptMessage):
    MESSAGE_ID = 0x0429
    _fields_ = _apt_short_message_fields([("chan_ident", ctypes.c_uint8)])

class _AptMsgGetStatusBits(_AptMessageHeader):
    MESSAGE_ID = 0x042A
    _fields_ = [("chan_ident",      ctypes.c_uint16),
                ("status_bits",     ctypes.c_uint32)]


# Build a table, mapping message ID to the corresponding Python class.
_apt_message_type_table: dict[int, Type["_AptMessage"]] = {
    _AptMsgIdentify.MESSAGE_ID:             _AptMsgIdentify,
    _AptMsgSetChanEnableState.MESSAGE_ID:   _AptMsgSetChanEnableState,
    _AptMsgReqChanEnableState.MESSAGE_ID:   _AptMsgReqChanEnableState,
    _AptMsgGetChanEnableState.MESSAGE_ID:   _AptMsgGetChanEnableState,
    _AptMsgHwReqInfo.MESSAGE_ID:            _AptMsgHwReqInfo,
    _AptMsgHwGetInfo.MESSAGE_ID:            _AptMsgHwGetInfo,
    _AptMsgReqPosCounter.MESSAGE_ID:        _AptMsgReqPosCounter,
    _AptMsgGetPosCounter.MESSAGE_ID:        _AptMsgGetPosCounter,
    _AptMsgSetVelParams.MESSAGE_ID:         _AptMsgSetVelParams,
    _AptMsgReqVelParams.MESSAGE_ID:         _AptMsgReqVelParams,
    _AptMsgGetVelParams.MESSAGE_ID:         _AptMsgGetVelParams,
    _AptMsgSetGenMoveParams.MESSAGE_ID:     _AptMsgSetGenMoveParams,
    _AptMsgReqGenMoveParams.MESSAGE_ID:     _AptMsgReqGenMoveParams,
    _AptMsgGetGenMoveParams.MESSAGE_ID:     _AptMsgGetGenMoveParams,
    _AptMsgSetHomeParams.MESSAGE_ID:        _AptMsgSetHomeParams,
    _AptMsgReqHomeParams.MESSAGE_ID:        _AptMsgReqHomeParams,
    _AptMsgGetHomeParams.MESSAGE_ID:        _AptMsgGetHomeParams,
    _AptMsgMoveHome.MESSAGE_ID:             _AptMsgMoveHome,
    _AptMsgMoveHomed.MESSAGE_ID:            _AptMsgMoveHomed,
    _AptMsgMoveRelative.MESSAGE_ID:         _AptMsgMoveRelative,
    _AptMsgMoveAbsolute.MESSAGE_ID:         _AptMsgMoveAbsolute,
    _AptMsgMoveCompleted.MESSAGE_ID:        _AptMsgMoveCompleted,
    _AptMsgMoveStop.MESSAGE_ID:             _AptMsgMoveStop,
    _AptMsgMoveStopped.MESSAGE_ID:          _AptMsgMoveStopped,
    _AptMsgReqStatusBits.MESSAGE_ID:        _AptMsgReqStatusBits,
    _AptMsgGetStatusBits.MESSAGE_ID:        _AptMsgGetStatusBits,
}


class VelocityParams(NamedTuple):
    """Velocity parameters for the K10CR1.

    Attributes:
        max_velocity:    Maximum velocity in degrees/second.
        acceleration:    Acceleration in degrees/second/second.
    """
    max_velocity: float
    acceleration: float


class HomeDirection(enum.IntEnum):
    """Possible values for the ``home_direction`` field in the homing parameters."""
    FORWARD = 1
    REVERSE = 2


class HomeLimitSwitch(enum.IntEnum):
    """Possible values for the ``limit_switch`` field in the homing parameters."""
    REVERSE = 1
    FORWARD = 4


class HomeParams(NamedTuple):
    """Homing parameters for the K10CR1.

    Attributes:
        home_direction:  Direction of moving to home (1 = forward, 2 = reverse).
        limit_switch:    Limit switch to use for homing (1 = reverse, 4 = forward).
        home_velocity:   Homing velocity in degrees/second.
        offset_distance: Distance of home postion from home limit switch (in degrees).
    """
    home_direction:     HomeDirection
    limit_switch:       HomeLimitSwitch
    home_velocity:      float
    offset_distance:    float


class MotorStatus(NamedTuple):
    """Status bits of the K10CR1 motorized stage.

    Some of the status bits do not seem to work with the K10CR1.

    Attributes:
        moving_forward:     True if the motor is moving in forward direction.
        moving_reverse:     True if the motor is moving in reverse direction.
                            It looks like `move_forward` and `move_reverse` are both
                            active when the stage is moving, regardless of the actual
                            direction of movement.
        jogging_forward:    True if the motor is jogging in forward direction.
        jogging_reverse:    True if the motor is jogging in reverse direction.
                            It looks like `jogging_reverse` is also active when jogging
                            in forward direction, while `jogging_forward` is never active.
        homing:             True if the motor is homing.
        homed:              True if homing has been completed.
        motion_error:       True if an excessive position error is detected.
        current_limit:      True if the motor current limit has been reached.
        channel_enabled:    True if the motor drive channel is enabled.
    """
    forward_limit:      bool
    reverse_limit:      bool
    moving_forward:     bool
    moving_reverse:     bool
    jogging_forward:    bool
    jogging_reverse:    bool
    homing:             bool
    homed:              bool
    tracking:           bool
    settled:            bool
    motion_error:       bool
    current_limit:      bool
    channel_enabled:    bool


class Thorlabs_K10CR1(QMI_Instrument):
    """Instrument driver for the Thorlabs K10CR1/M motorized rotational mount."""

    # Number of microsteps per degree of rotation.
    MICROSTEPS_PER_DEGREE = 409600.0 / 3.0

    # Internal velocity setting for 1 degree/second.
    VELOCITY_FACTOR = 7329109.0

    # Internal acceleration setting for 1 degree/second/second.
    ACCELERATION_FACTOR = 1502.0

    def __init__(self, context: QMI_Context, name: str, transport: str) -> None:
        """Initialize driver.

        The motorized mount presents itself as a USB serial port.
        The transport descriptor should refer to the serial port device,
        e.g. "serial:/dev/ttyUSB1"

        Parameters:
            name: Name for this instrument instance.
            transport: Transport descriptor to access the instrument.
        """
        super().__init__(context, name)
        self._transport = create_transport(transport, default_attributes={"baudrate": 115200, "rtscts": True})

        # Expect a reply to normal commands within 1 second.
        self._reply_timeout = 1.0

        # The APT protocol supports multiple channels.
        # The K10CR1 only uses channel 1.
        self._channel = 1

    def _read_message(self, timeout: float) -> _AptMessage:
        """Read and decode a binary message from the instrument."""

        # Read message header.
        data = self._transport.read(nbytes=6, timeout=timeout)

        # Decode message header.
        hdr = _AptMessageHeader.from_buffer_copy(data)

        # Long APT messages are identified by bit 7 in the destination field.
        if (hdr.dest & 0x80) != 0:
            # This is a long APT message (header + data). Read the additional data.
            try:
                # Since we already received a partial message, the timeout
                # only needs to account for the time it takes to receive
                # the payload data. (The instrument will probably transmit
                # the entire message as fast as possible).
                # 50 ms should be more than enough.
                data += self._transport.read(nbytes=hdr.data_length, timeout=0.050)
            except QMI_TimeoutException:
                # Discard pending data after receiving a partial message.
                self._transport.discard_read()
                raise QMI_InstrumentException(
                    "Received partial message (message_id=0x{:04x}, data_length={})".format(
                        hdr.message_id, hdr.data_length
                    )
                )

        # Decode the complete message.
        message_type = _apt_message_type_table.get(hdr.message_id)
        if message_type is None:
            # Discard pending data after receiving a bad message.
            self._transport.discard_read()
            raise QMI_InstrumentException(
                "Received unknown message id 0x{:04x} from instrument".format(hdr.message_id)
            )

        if len(data) != ctypes.sizeof(message_type):
            # Discard pending data after receiving a bad message.
            self._transport.discard_read()
            raise QMI_InstrumentException(
                ("Received incorrect message length for message id 0x{:04x} "
                + "(got {} bytes while expecting {} bytes)").format(
                    hdr.message_id, len(data), ctypes.sizeof(message_type)
                )
            )

        # Decode received message.
        return message_type.from_buffer_copy(data)

    def _wait_message(self, message_type: Type[_AptMessage], timeout: float) -> _AptMessage:
        """Wait for a specific message type from the instrument.

        Any other (valid) messages received from the instrument will be discarded.

        Parameters:
            message_type:   Type of message to wait for.
            timeout:        Maximum time to wait for the message in seconds.

        Returns:
            The received message.

        Raises:
            QMI_TimeoutException: If the expected message is not received within the timeout.
            QMI_InstrumentException: If an invalid message is received.
        """

        end_time = time.monotonic() + timeout
        while True:

            # Read next message from instrument.
            tmo = max(end_time - time.monotonic(), 0)
            msg = self._read_message(timeout=tmo)

            if isinstance(msg, message_type):
                # Got the expected message.
                return msg

            # Discard message and continue waiting.
            _logger.debug("[%s] Ignoring message %s (message_id=0x%04x)",
                          self._name, type(msg).__name__, msg.message_id)

            if time.monotonic() > end_time:
                raise QMI_TimeoutException(f"Expected message type {message_type} not received.")

    def _send_message(self, msg: _AptMessage) -> None:
        """Encode and send a binary message to the instrument."""
        # Before sending a new command, do a non-blocking read to consume
        # a potential old message from the instrument.
        # This prevents a buildup of unhandled notification messages from
        # the instrument after several move_XXX() commands.
        # Suppress the timeout exception if no old message is in the buffer.
        with contextlib.suppress(QMI_TimeoutException):
            pending_msg = self._read_message(timeout=0.0)
            _logger.debug("[%s] Pending message %s (message_id=0x%04x)",
                          self._name, type(pending_msg).__name__, pending_msg.message_id)

        # Now send the new command.
        self._transport.write(bytes(msg))

    def _check_k10cr1(self) -> None:
        """Check that the connected device is a Thorlabs K10CR1.

        Raises:
            QMI_InstrumentException: If not connected to a K10CR1 device.
            QMI_TimeoutException: If the instrument does not answer our request.
        """

        # Send request message.
        req_msg = _AptMsgHwReqInfo.create()
        self._send_message(req_msg)

        # Receive reply message.
        reply_msg = self._wait_message(_AptMsgHwGetInfo, timeout=self._reply_timeout)

        # Check that this is a K10CR1 device.
        model_str = reply_msg.model_number.decode("iso8859-1")
        if model_str != "K10CR1":
            raise QMI_InstrumentException(
                "Driver only supports K10CR1 but instrument identifies as {!r}".format(model_str)
            )

    @rpc_method
    def open(self) -> None:
        _logger.info("[%s] Opening connection to instrument", self._name)
        self._check_is_closed()
        self._transport.open()

        try:
            # Wait and discard partial data, as recommended by the APT protocol documentation.
            time.sleep(0.050)
            self._transport.discard_read()

            # NOTE: The APT doc says we should wait another 50 ms and then enable RTS.
            #   We can not easily do this because our transport layer does not support explicit set/clear RTS.
            #   Skip this step for now; we can try it later if there are communication problems.

            # Check that this device is a K10CR1 motor.
            # Otherwise we should not talk to it, since we don't want to send
            # inappropriate commands to some unsupported device.
            self._check_k10cr1()

        except Exception:
            # Close the transport if an error occurred during initialization
            # of the instrument.
            self._transport.close()
            raise

        # Mark this instrument as open.
        super().open()

    @rpc_method
    def close(self) -> None:
        _logger.info("[%s] Closing connection to instrument", self._name)
        super().close()
        self._transport.close()

    @rpc_method
    def get_idn(self) -> QMI_InstrumentIdentification:
        """Read instrument type and version and return QMI_InstrumentIdentification instance."""
        self._check_is_open()

        # Send request message.
        req_msg = _AptMsgHwReqInfo.create()
        self._send_message(req_msg)

        # Receive reply message.
        reply_msg = self._wait_message(_AptMsgHwGetInfo, timeout=self._reply_timeout)

        return QMI_InstrumentIdentification(
            vendor="Thorlabs",
            model=reply_msg.model_number.decode("iso8859-1"),
            serial=reply_msg.serial_number,
            version="{}.{}.{}".format(reply_msg.fw_version[2], reply_msg.fw_version[1], reply_msg.fw_version[0])
        )

    @rpc_method
    def get_motor_status(self) -> MotorStatus:
        """Return the motor status bits."""
        self._check_is_open()

        # Send request message.
        req_msg = _AptMsgReqStatusBits.create(chan_ident=self._channel)
        self._send_message(req_msg)

        # Receive reply message.
        reply_msg = self._wait_message(_AptMsgGetStatusBits, timeout=self._reply_timeout)

        # Decode motor status bits.
        status_bits = reply_msg.status_bits
        return MotorStatus(
            forward_limit=((status_bits & 0x01) != 0),
            reverse_limit=((status_bits & 0x02) != 0),
            moving_forward=((status_bits & 0x10) != 0),
            moving_reverse=((status_bits & 0x20) != 0),
            jogging_forward=((status_bits & 0x40) != 0),
            jogging_reverse=((status_bits & 0x80) != 0),
            homing=((status_bits & 0x200) != 0),
            homed=((status_bits & 0x400) != 0),
            tracking=((status_bits & 0x1000) != 0),
            settled=((status_bits & 0x2000) != 0),
            motion_error=((status_bits & 0x4000) != 0),
            current_limit=((status_bits & 0x01000000) != 0),
            channel_enabled=((status_bits & 0x80000000) != 0)
        )

    @rpc_method
    def identify(self) -> None:
        """Make the yellow LED on the instrument flash for a few seconds."""
        self._check_is_open()

        # Send command message.
        req_msg = _AptMsgIdentify.create(chan_ident=self._channel)
        self._send_message(req_msg)

    @rpc_method
    def get_chan_enable_state(self) -> bool:
        """Return the enable state of the motor drive channel."""
        self._check_is_open()

        # Send request message.
        req_msg = _AptMsgReqChanEnableState.create(chan_ident=self._channel)
        self._send_message(req_msg)

        # Receive reply message.
        reply_msg = self._wait_message(_AptMsgGetChanEnableState, timeout=self._reply_timeout)
        return reply_msg.enable_state == 0x01

    @rpc_method
    def set_chan_enable_state(self, enable: bool) -> None:
        """Enable or disable the motor drive channel.

        When the drive channel is disabled, the motor will not hold its
        position and can be moved through external force.

        The drive channel is by default enabled at power-up.
        """
        self._check_is_open()

        # Send command message.
        enable_state = 0x01 if enable else 0x02
        req_msg = _AptMsgSetChanEnableState.create(chan_ident=self._channel, enable_state=enable_state)
        self._send_message(req_msg)

    @rpc_method
    def get_absolute_position(self) -> float:
        """Return the absolute position of the stage in degrees.

        After power-up, the absolute position of the stage will be
        unknown until the stage is homed. If the stage has not yet
        been homed, this function will return the current position
        relative to the position at power-up.

        Note that the absolute position can exceed the range of
        -360 .. +360 degrees when the stage has rotated a full turn.
        The absolute position counter overflows after 43 full turns
        and may then return incorrect results.
        """
        self._check_is_open()

        # Send request message.
        req_msg = _AptMsgReqPosCounter.create(chan_ident=self._channel)
        self._send_message(req_msg)

        # Receive reply message.
        reply_msg = self._wait_message(_AptMsgGetPosCounter, timeout=self._reply_timeout)
        return reply_msg.position / self.MICROSTEPS_PER_DEGREE

    @rpc_method
    def get_velocity_params(self) -> VelocityParams:
        """Return the current maximum velocity and acceleration."""

        self._check_is_open()

        # Send request message.
        req_msg = _AptMsgReqVelParams.create(chan_ident=self._channel)
        self._send_message(req_msg)

        # Receive reply message.
        reply_msg = self._wait_message(_AptMsgGetVelParams, timeout=self._reply_timeout)
        return VelocityParams(
            max_velocity=(reply_msg.max_vel / self.VELOCITY_FACTOR),
            acceleration=(reply_msg.accel / self.ACCELERATION_FACTOR)
        )

    @rpc_method
    def set_velocity_params(self, max_velocity: float, acceleration: float) -> None:
        """Set the maximum velocity and acceleration.

        These settings will be applied for subsequent absolute and relative moves.

        Parameters:
            max_velocity: Maximum velocity in degrees/second (max 10).
            acceleration: Acceleration in degree/second/second (max 20).
        """

        if max_velocity <= 0 or max_velocity > 10:
            raise ValueError("Invalid range for max_velocity")
        if acceleration <= 0 or acceleration > 20:
            raise ValueError("Invalid range for acceleration")

        self._check_is_open()

        # Send command message.
        max_vel = int(round(max_velocity * self.VELOCITY_FACTOR))
        accel = int(round(acceleration * self.ACCELERATION_FACTOR))
        req_msg = _AptMsgSetVelParams.create(
            chan_ident=self._channel,
            min_vel=0,
            accel=accel,
            max_vel=max_vel
        )
        self._send_message(req_msg)

    @rpc_method
    def get_backlash_distance(self) -> float:
        """Return the backlash distance in degrees."""

        self._check_is_open()

        # Send request message.
        req_msg = _AptMsgReqGenMoveParams.create(chan_ident=self._channel)
        self._send_message(req_msg)

        # Receive reply message.
        reply_msg = self._wait_message(_AptMsgGetGenMoveParams, timeout=self._reply_timeout)
        return reply_msg.backlash_dist / self.MICROSTEPS_PER_DEGREE

    @rpc_method
    def set_backlash_distance(self, backlash: float) -> None:
        """Set the backlash distance.

        The backlash distance is used when the stage moves in reverse direction.
        While moving in reverse direction, the stage moves past its target position
        by the backlash distance, then moves back to the target in forward direction.

        Parameters:
            backlash: Backlash distance in degrees, or 0 to disable.
        """

        # Convert distance to microsteps and check that the value fits in a 32-bit signed integer.
        raw_dist = int(round(backlash * self.MICROSTEPS_PER_DEGREE))
        if abs(raw_dist) >= 2**31:
            raise ValueError("Backlash distance out of valid range")

        self._check_is_open()

        # Send command message.
        req_msg = _AptMsgSetGenMoveParams.create(chan_ident=self._channel, backlash_dist=raw_dist)
        self._send_message(req_msg)

    @rpc_method
    def get_home_params(self) -> HomeParams:
        """Return the homing parameters."""

        self._check_is_open()

        # Send request message.
        req_msg = _AptMsgReqHomeParams.create(chan_ident=self._channel)
        self._send_message(req_msg)

        # Receive reply message.
        reply_msg = self._wait_message(_AptMsgGetHomeParams, timeout=self._reply_timeout)
        return HomeParams(
            home_direction=HomeDirection(reply_msg.home_dir),
            limit_switch=HomeLimitSwitch(reply_msg.limit_switch),
            home_velocity=(reply_msg.home_velocity / self.VELOCITY_FACTOR),
            offset_distance=(reply_msg.offset_dist / self.MICROSTEPS_PER_DEGREE)
        )

    @rpc_method
    def set_home_params(
        self,
        home_direction: HomeDirection,
        limit_switch: HomeLimitSwitch,
        home_velocity: float,
        offset_distance: float
    ) -> None:
        """Set the homing parameters.

        WARNING: The K10CR1 manual recommends that these settings should
        not be adjusted from the factory defaults.

        Parameters:
            home_direction:     Direction of moving to home (should be HomeDirection.REVERSE).
            limit_switch:       Limit switch to use for homing (should be HomeLimitSwitch.REVERSE).
            home_velocity:      Homing velocity in degrees/second (max 5).
            offset_distance:    Distance of home position from home limit switch (in degrees).
        """

        if home_direction not in (HomeDirection.FORWARD, HomeDirection.REVERSE):
            raise ValueError("Invalid value for home_direction")
        if limit_switch not in (HomeLimitSwitch.FORWARD, HomeLimitSwitch.REVERSE):
            raise ValueError("Invalid value for limit_switch")
        if home_velocity <= 0 or home_velocity > 5:
            raise ValueError("Invalid range for home_velocity")

        # Convert distance to microsteps and check that the value fits in a 32-bit signed integer.
        raw_dist = int(round(offset_distance * self.MICROSTEPS_PER_DEGREE))
        if abs(raw_dist) >= 2**31:
            raise ValueError("Invalid range for offset_distance")

        self._check_is_open()

        # Send command message.
        raw_velocity = int(round(home_velocity * self.VELOCITY_FACTOR))
        req_msg = _AptMsgSetHomeParams.create(
            chan_ident=self._channel,
            home_dir=int(home_direction),
            limit_switch=int(limit_switch),
            home_velocity=raw_velocity,
            offset_dist=raw_dist
        )
        self._send_message(req_msg)

    @rpc_method
    def move_stop(self, immediate_stop: bool = False) -> None:
        """Stop any ongoing move operation.

        This function returns immediately while the stage may still be moving.
        Use ``wait_move_complete()`` to wait until the stage has stopped.

        Parameters:
            immediate_stop: True to stop abruptly, False to perform a profiled stop.
        """
        self._check_is_open()

        # Send command message.
        stop_mode = 0x01 if immediate_stop else 0x02
        req_msg = _AptMsgMoveStop.create(chan_ident=self._channel, stop_mode=stop_mode)
        self._send_message(req_msg)

    @rpc_method
    def move_home(self) -> None:
        """Start a homing move.

        This function returns immediately while homing is still in progress.
        Use ``wait_move_complete()`` to wait until the move is finished.
        """
        self._check_is_open()

        # Send command message.
        req_msg = _AptMsgMoveHome.create(chan_ident=self._channel)
        self._send_message(req_msg)

    @rpc_method
    def move_relative(self, distance: float) -> None:
        """Start a relative move.

        This function returns immediately while homing is still in progress.
        Use ``wait_move_complete()`` to wait until the move is finished.

        Parameters:
            distance: Relative move distance in degrees.
        """

        # Convert distance to microsteps and check that the value fits in a 32-bit signed integer.
        raw_dist = int(round(distance * self.MICROSTEPS_PER_DEGREE))
        if abs(raw_dist) >= 2**31:
            raise ValueError("Relative distance out of valid range")

        self._check_is_open()

        # Send command message.
        req_msg = _AptMsgMoveRelative.create(chan_ident=self._channel, rel_dist=raw_dist)
        self._send_message(req_msg)

    @rpc_method
    def move_absolute(self, position: float) -> None:
        """Start a move to an absolute position.

        This function returns immediately while homing is still in progress.
        Use ``wait_move_complete()`` to wait until the move is finished.

        After power-up, the absolute position of the stage will be
        unknown until the stage is homed. If the stage has not yet
        been homed, the position parameter of this function will
        be interpreted relative to the position at power-up.

        Note that the stage supports absolute positions above 360 degrees
        and will not automatically reduce such positions module 360.
        It is therefore possible that the stage must rotate more than
        a full turn in order to reach a specific absolute position.

        Parameters:
            position: Absolute target position in degrees.
        """

        # Convert distance to microsteps and check that the value fits in a 32-bit signed integer.
        raw_pos = int(round(position * self.MICROSTEPS_PER_DEGREE))
        if abs(raw_pos) >= 2**31:
            raise ValueError("Absolute position out of valid range")

        self._check_is_open()

        # Send command message.
        req_msg = _AptMsgMoveAbsolute.create(chan_ident=self._channel, abs_position=raw_pos)
        self._send_message(req_msg)

    @rpc_method
    def wait_move_complete(self, timeout: float) -> None:
        """Wait until the motor has stopped moving.

        If the motor is not currently moving, this function returns immediately.
        Otherwise, this function wait (blocks) until the motor has stopped moving
        or until the specified timeout expires.

        Parameters:
            timeout: Maximum time to wait in seconds.

        Raises:
            QMI_TimeoutException: If the motor is still moving when the timeout expires.
        """
        self._check_is_open()
        _logger.debug("[%s] wait_move_complete(%f)", self._name, timeout)

        end_time = time.monotonic() + timeout

        while True:
            # Read the motor status to see if we are moving.
            status = self.get_motor_status()
            if (not (status.moving_forward
                     or status.moving_reverse
                     or status.jogging_forward
                     or status.jogging_reverse
                     or status.homing)):
                # Not moving anymore. We are done here.
                _logger.debug("[%s] Not moving", self._name)
                return

            # Check if we reached the timeout.
            time_left = end_time - time.monotonic()
            if time_left <= 0:
                raise QMI_TimeoutException("Timeout while waiting for end of move")

            # Wait for a short while, or until the motor sends a new message.
            try:
                msg = self._read_message(timeout=min(0.5, time_left))
            except QMI_TimeoutException:
                # No message from the motor.
                # This is normal and expected if the motor is still moving.
                # Go around the loop again and poll the motor status.
                continue

            # Any message from the motor is most likely an announcement that
            # the move has ended. Whatever the actual message, go around
            # the loop and poll the motor status again.
            _logger.debug("[%s] Ignoring message %s (message_id=0x%04x)",
                          self._name, type(msg).__name__, msg.message_id)



"""Mapping of error number to error message for Newport TLB670x laser controller."""
ERROR_MESSAGES = {
    -4: "NO PARAMETER SPECIFIED",
    -3: "LASER HEAD DISCONNECTED",
    -2: "COMMAND NOT VALID",
    0: "NO ERROR",
    # No error exists in the error buffer
    116: "SYNTAX ERROR",
    # This error is generated when the instrument receives a command that cannot be processed.
    # Some typical causes:
    # a. Using ASCII characters outside of a string constant that are not defined by the command language syntax.
    # b. Missing space between a set command and parameter.
    # c. Missing “?” character in case of query
    126: "WRONG NUM OF PARAMS",
    # This error is generated when the instrument is unable to process a command due to a mismatch between the number
    # of parameters received and the number of parameters required for the command.
    201: "VALUE OUT OF RANGE",
    # This error is generated when the instrument is unable to process a command because the parameter value received
    # is out of range of the acceptable values for the command.
    424: "TEC OVER TEMP",
    # The temperature control subsystem output has been turned OFF because the internal heatsink temperature has
    # exceeded safe levels. Please allow the unit to cool down, and reset the controller by turning it OFF and back ON.
    501: "LDD INTERLOCK FLT",
    # The Laser Diode Driver subsystem output has been turned OFF because of remote interlock (BNC connector on rear
    # panel) assertion.
    502: "LDD CURRENT LIMIT",
    # The Laser Diode Driver subsystem output has been turned OFF because of actual current exceeding the current limit
    # set point.
    514: "LDD MODE CHANGE",
    # The Laser Diode Driver subsystem output has been turned OFF because a mode change was commanded using
    # “SOURce:CPOWer” command or through front-panel interface.
    524: "LDD OVER TEMP",
    # The Laser Diode Driver subsystem output has been turned OFF because the internal heatsink temperature has
    # exceeded safe levels. Please allow the unit to cool down before turning the output ON again.
    803: "WAVE OVER TEMP",
    # The Wavelength control subsystem output has been turned OFF because the internal heatsink temperature
    # has exceeded safe levels. Please allow the unit to cool down before turning the wavelength tracking ON again.
    802: "WAVE FOLLOW ERROR",
    # Wavelength tracking has been turned OFF because the difference between actual wavelength and wavelength
    # setpoint exceeded 2nm. Please reduce the scan speed setting and turn wavelength tracking ON again.
    804: "PIEZO OVER TEMP",
    # The piezo control subsystem setpoint has been changed to 0% because the internal heatsink temperature has
    # exceeded safe levels. Please allow the unit to cool down before changing the piezo setpoint again.
    805: "CMD NOT SUPPORTED",
    # This error is generated when the command received by the controller is not supported by laser head connected to
    # it
    902: "KEY ENABLE OFF",
    # This error is generated when user tries to turn the Laser Diode Driver subsystem output ON with key switch in
    # OFF position.
    919: "LASER DISCONNECTED",
    # This error is generated when user tries to turn the Laser Diode Driver subsystem output ON without having a
    # laser head connected to the controller.
    920: "LASER EEPROM ERROR",
    # This error is generated when user tries to connect a laser head to the controller using an incompatible cable.
}



""" Testcase of the Tenma series 72 power supply units."""
import logging
import unittest
from unittest.mock import call, patch

from qmi.core.exceptions import QMI_InstrumentException
from qmi.core.transport import QMI_UdpTransport
from qmi.instruments.tenma import Tenma72_2550, Tenma72_13350
from qmi.instruments.tenma.psu_72 import Tenma72_Base

from tests.patcher import PatcherQmiContext as QMI_Context


class TestTenma72_BaseInit(unittest.TestCase):
    """ Testcase of the Tenma72_Base class initialization"""

    def setUp(self):
        # Add patches
        patcher = patch('qmi.instruments.tenma.psu_72.create_transport', spec=QMI_UdpTransport)
        self._transport_mock = patcher.start().return_value
        self.addCleanup(patcher.stop)

    def test_init(self):
        """See that __init__ excepts when invalid type of transport is given."""
        # Arrange
        ok_transport = "serial:/dev/ttyS0"
        ok_transport_2 = "udp:123.45.67.8"
        nok_transport = "tcp:123.45.67.8:1234"  # TCP cannot be used
        # Act
        Tenma72_Base(
            QMI_Context("Test_tenma_base_init"), "tenma_base_ok", ok_transport
        )
        Tenma72_Base(
            QMI_Context("Test_tenma_base_init2"), "tenma_base_ok2", ok_transport_2
        )
        # Assert
        with self.assertRaises(QMI_InstrumentException):
            Tenma72_Base(
                QMI_Context("Test_tenma_base_init_nok"), "tenma_base_nok", nok_transport
            )


class TestTenma72_Base(unittest.TestCase):
    """ Testcase of the Tenma72_Base class """

    def setUp(self):
        # Add patches
        patcher = patch('qmi.instruments.tenma.psu_72.create_transport', spec=QMI_UdpTransport)
        self._transport_mock = patcher.start().return_value
        self.addCleanup(patcher.stop)
        # Make DUT
        self.psu: Tenma72_Base = Tenma72_Base(QMI_Context("TestSiglentTenma72_Base"), Tenma72_Base, "udp:123.45.67.8")
        self.psu.open()

    def tearDown(self):
        self.psu.close()

    def test_get_idn(self):
        """ Test case for `get_idn()` function. """
        # arrange
        expected_vendor = "TENMA"
        expected_model = "72-2345"
        expected_serial = "1231345"
        expected_version = "V2.0"
        self._transport_mock.read_until_timeout.return_value = "TENMA 72-2345 SN:1231345 V2.0".encode("ascii")
        expected_calls = [call.read_until_timeout(Tenma72_Base.BUFFER_SIZE, 0.2)]  # 0.2 = default timeout
        # act
        idn = self.psu.get_idn()
        # assert
        self.assertEqual(expected_calls, self._transport_mock.read_until_timeout.call_args_list)
        self.assertEqual(expected_vendor, idn.vendor)
        self.assertEqual(expected_model, idn.model)
        self.assertEqual(expected_serial, idn.serial)
        self.assertEqual(expected_version, idn.version)

    def test_get_idn_no_reply_excepts(self):
        """ Test case for `get_idn()` function. """
        # arrange
        self._transport_mock.read_until_timeout.return_value = "".encode("ascii")
        # act & assert
        with self.assertRaises(QMI_InstrumentException):
            self.psu.get_idn()

    def test_get_status_errs(self):
        """Test case for not implemented get_status() function."""
        with self.assertRaises(NotImplementedError):
            self.psu.get_status()

    def test_read_current(self):
        """Test base read_current(...) method with a channel number and without."""
        # Arrange
        cmd_1 = "ISET?"  # No channel
        ch_2 = 2
        cmd_2 = f"ISET{ch_2}?"
        expected_current = 0.123
        self._transport_mock.read_until_timeout.return_value = f"{expected_current:.5f}".encode("ascii")
        # Act
        current = self.psu.get_current()  # No channel
        # Assert
        self.assertEqual(expected_current, current)
        self._transport_mock.write.assert_called_once_with(bytes(cmd_1, "ascii"))
        self._transport_mock.write.reset_mock()
        # Act
        current = self.psu.get_current(ch_2)
        # Assert
        self.assertEqual(expected_current, current)
        self._transport_mock.write.assert_called_once_with(bytes(cmd_2, "ascii"))

    def test_read_voltage(self):
        """Test base read_voltage(...) method with a channel number and without."""
        # Arrange
        cmd_1 = "VSET?"  # No channel
        ch_2 = 2
        cmd_2 = f"VSET{ch_2}?"
        expected_voltage = 1.230
        self._transport_mock.read_until_timeout.return_value = f"{expected_voltage:.5f}".encode("ascii")
        # Act
        voltage = self.psu.get_voltage()  # No channel
        # Assert
        self.assertEqual(expected_voltage, voltage)
        self._transport_mock.write.assert_called_once_with(bytes(cmd_1, "ascii"))
        self._transport_mock.write.reset_mock()
        # Act
        voltage = self.psu.get_voltage(ch_2)
        # Assert
        self.assertEqual(expected_voltage, voltage)
        self._transport_mock.write.assert_called_once_with(bytes(cmd_2, "ascii"))

    def test_enable_disable_output(self):
        """Test enabling and disabling output. At base class sending works but receiving excepts"""
        # Act
        with self.assertRaises(NotImplementedError):
            self.psu.enable_output(True)

        with self.assertRaises(NotImplementedError):
            self.psu.enable_output(False)


class TestTenma72_2550(unittest.TestCase):
    """ Testcase of the TestTenma72_2550 psu """

    def setUp(self):
        # Add patches
        patcher = patch('qmi.instruments.tenma.psu_72.create_transport', spec=QMI_UdpTransport)
        self._transport_mock = patcher.start().return_value
        self.addCleanup(patcher.stop)
        # Make DUT
        self.psu: Tenma72_2550 = Tenma72_2550(QMI_Context("TestSiglentTenma72_2550"), Tenma72_2550, "serial:COM0")
        self.psu.open()

    def tearDown(self):
        self.psu.close()

    def test_get_idn(self):
        """ Test case for `get_idn()` function. """
        # arrange
        expected_vendor = "TENMA"
        expected_model = "72-2550"
        expected_serial = "1231345"
        expected_version = "V2.0"
        self._transport_mock.read_until_timeout.return_value = "TENMA 72-2550 SN:1231345 V2.0".encode("ascii")
        expected_calls = [call.read_until_timeout(Tenma72_2550.BUFFER_SIZE, 0.2)]  # 0.2 = default timeout
        # act
        idn = self.psu.get_idn()
        # assert
        self.assertEqual(expected_calls, self._transport_mock.read_until_timeout.call_args_list)
        self.assertEqual(expected_vendor, idn.vendor)
        self.assertEqual(expected_model, idn.model)
        self.assertEqual(expected_serial, idn.serial)
        self.assertEqual(expected_version, idn.version)

    def test_get_status(self):
        """Test case for get_status() function."""
        # arrange
        responses = [chr(0x01), chr(0x02), chr(0x03), chr(0x07), chr(0x0B), chr(0x0C), chr(0x40)]
        self._transport_mock.read.side_effect = responses
        expected_statuses = [
            {"Ch1Mode": "C.V", "Ch2Mode": "C.C", "Tracking": "Independent", "OutputEnabled": False},
            {"Ch1Mode": "C.C", "Ch2Mode": "C.V", "Tracking": "Independent", "OutputEnabled": False},
            {"Ch1Mode": "C.V", "Ch2Mode": "C.V", "Tracking": "Independent", "OutputEnabled": False},
            {"Ch1Mode": "C.V", "Ch2Mode": "C.V", "Tracking": "Tracking Series", "OutputEnabled": False},
            {"Ch1Mode": "C.V", "Ch2Mode": "C.V", "Tracking": "Tracking Parallel", "OutputEnabled": False},
            {"Ch1Mode": "C.C", "Ch2Mode": "C.C", "Tracking": "Unknown", "OutputEnabled": False},
            {"Ch1Mode": "C.C", "Ch2Mode": "C.C", "Tracking": "Independent", "OutputEnabled": True}
        ]
        # Act
        for expected in expected_statuses:
            status = self.psu.get_status()
            # Assert
            self.assertDictEqual(expected, status)
            self._transport_mock.read.assert_called_once_with(1, 1)
            self._transport_mock.reset_mock()

    def test_set_current(self):
        """Set current within allowed range."""
        # Arrange
        allowed_currents = [Tenma72_2550.MAX_CURRENT, 0]
        channels = [None, 2]
        expected_calls = [
            call.write(f"ISET:{allowed_currents[0]:.3f}".encode("ascii")),
            call.write(f"ISET{channels[1]}:{allowed_currents[1]:.3f}".encode("ascii"))
        ]
        # Act
        for e, current in enumerate(allowed_currents):
            self.psu.set_current(current, channels[e])

        # Assert
        self._transport_mock.write.assert_has_calls(expected_calls)

    def test_set_current_excepts(self):
        """See that values out-of-range cause an exception."""
        # Arrange
        disallowed_currents = [Tenma72_2550.MAX_CURRENT + 0.1, -0.1]
        # Act
        for current in disallowed_currents:
            with self.assertRaises(ValueError):
                self.psu.set_current(current)

        self._transport_mock.write.assert_not_called()

    def test_set_voltage(self):
        """Set voltage within allowed range."""
        # Arrange
        allowed_voltages = [Tenma72_2550.MAX_VOLTAGE, 0]
        channels = [None, 2]
        expected_calls = [
            call.write(f"VSET:{allowed_voltages[0]:.3f}".encode("ascii")),
            call.write(f"VSET{channels[1]}:{allowed_voltages[1]:.3f}".encode("ascii"))
        ]
        # Act
        for e, voltage in enumerate(allowed_voltages):
            self.psu.set_voltage(voltage, channels[e])

        self._transport_mock.write.assert_has_calls(expected_calls)

    def test_set_voltage_excepts(self):
        """See that values out-of-range cause an exception."""
        # Arrange
        disallowed_voltages = [Tenma72_2550.MAX_VOLTAGE + 0.1, -0.1]
        # Act
        for voltage in disallowed_voltages:
            with self.assertRaises(ValueError):
                self.psu.set_voltage(voltage)

        self._transport_mock.write.assert_not_called()

    def test_enable_disable_output(self):
        """Test enabling and disabling output. At implemented class this should work."""
        # Arrange
        enable_cmd = "OUT1"
        disable_cmd = "OUT0"
        self._transport_mock.read.side_effect = [chr(0x40), chr(0x00)]
        expected_enable_calls = [
            call(enable_cmd.encode("ascii")),
            call(b"STATUS?")
        ]
        expected_disable_calls = [
            call(disable_cmd.encode("ascii")),
            call(b"STATUS?")
        ]
        # Act
        self.psu.enable_output(True)

        # Assert
        self._transport_mock.write.assert_has_calls(expected_enable_calls)
        self._transport_mock.write.reset_mock()

        # Act
        self.psu.enable_output(False)

        # Assert
        self._transport_mock.write.assert_has_calls(expected_disable_calls)

    def test_enable_disable_output_excepts(self):
        """Test enabling or disabling output raises an exception when the check value is not the same as input."""
        # Suppress logging.
        logging.getLogger("qmi.instruments.tenma.psu_72").setLevel(logging.CRITICAL)

        # Arrange
        enable_cmd = "OUT1"
        disable_cmd = "OUT0"
        self._transport_mock.read.side_effect = [chr(0x40), chr(0x00)]
        expected_enable_calls = [
            call(enable_cmd.encode("ascii")),
            call(b"STATUS?")
        ]
        expected_disable_calls = [
            call(disable_cmd.encode("ascii")),
            call(b"STATUS?")
        ]
        # Act
        with self.assertRaises(QMI_InstrumentException):
            self.psu.enable_output(False)

        # Assert
        self._transport_mock.write.assert_has_calls(expected_disable_calls)
        self._transport_mock.write.reset_mock()

        # Act
        with self.assertRaises(QMI_InstrumentException):
            self.psu.enable_output(True)

        # Assert
        self._transport_mock.write.assert_has_calls(expected_enable_calls)

        logging.getLogger("qmi.instruments.tenma.psu_72").setLevel(logging.NOTSET)


class TestTenma72_13350(unittest.TestCase):
    """ Testcase of the TestTenma72_13350 psu """

    def setUp(self):
        # Add patches
        patcher = patch('qmi.instruments.tenma.psu_72.create_transport', spec=QMI_UdpTransport)
        self._transport_mock = patcher.start().return_value
        self.addCleanup(patcher.stop)
        # Make DUT
        self.psu: Tenma72_13350 = Tenma72_13350(QMI_Context("TestSiglentTenma72_13350"), Tenma72_13350, "udp:not_parsed")
        self.psu.open()

    def tearDown(self):
        self.psu.close()

    def test_get_idn(self):
        """ Test case for `get_idn()` function. """
        # arrange
        expected_vendor = "TENMA"
        expected_model = "72-13350"
        expected_serial = "1231345"
        expected_version = "V2.0"
        self._transport_mock.read_until_timeout.return_value = "TENMA 72-13350 SN:1231345 V2.0".encode("ascii")
        expected_calls = [call.read_until_timeout(Tenma72_13350.BUFFER_SIZE, 0.2)]  # 0.2 = default timeout
        # act
        idn = self.psu.get_idn()
        # assert
        self.assertEqual(expected_calls, self._transport_mock.read_until_timeout.call_args_list)
        self.assertEqual(expected_vendor, idn.vendor)
        self.assertEqual(expected_model, idn.model)
        self.assertEqual(expected_serial, idn.serial)
        self.assertEqual(expected_version, idn.version)

    def test_get_status(self):
        """Test case for get_status() function."""
        # arrange
        responses = [b"\x01\n", b"\x02\n", b"\x04\n", b"\x10\n", b" \n"]  # last is chr(0x20)
        self._transport_mock.read.side_effect = responses
        expected_statuses = [
            {"ChannelMode": "C.V", "OutputEnabled": False, "V/C priority": "Voltage priority", "Beep": False,
             "Lock": False},
            {"ChannelMode": "C.C", "OutputEnabled": True, "V/C priority": "Voltage priority", "Beep": False,
             "Lock": False},
            {"ChannelMode": "C.C", "OutputEnabled": False, "V/C priority": "Current priority", "Beep": False,
             "Lock": False},
            {"ChannelMode": "C.C", "OutputEnabled": False, "V/C priority": "Voltage priority", "Beep": True,
             "Lock": False},
            {"ChannelMode": "C.C", "OutputEnabled": False, "V/C priority": "Voltage priority", "Beep": False,
             "Lock": True}
        ]
        # Act
        for expected in expected_statuses:
            status = self.psu.get_status()
            # Assert
            self.assertDictEqual(expected, status)
            self._transport_mock.read.assert_called_once_with(2, 2)
            self._transport_mock.reset_mock()

    def test_set_current(self):
        """Set current within allowed range."""
        # Arrange
        allowed_currents = [Tenma72_13350.MAX_CURRENT, 0]
        channels = [None, 2]
        expected_calls = [
            call.write(f"ISET:{allowed_currents[0]:.3f}\n".encode("ascii")),
            call.write(f"ISET{channels[1]}:{allowed_currents[1]:.3f}\n".encode("ascii"))
        ]
        # Act
        for e, current in enumerate(allowed_currents):
            self.psu.set_current(current, channels[e])

        # Assert
        self._transport_mock.write.assert_has_calls(expected_calls)

    def test_set_current_excepts(self):
        """See that values out-of-range cause an exception."""
        # Arrange
        disallowed_currents = [Tenma72_13350.MAX_CURRENT + 0.1, -0.1]
        # Act
        for current in disallowed_currents:
            with self.assertRaises(ValueError):
                self.psu.set_current(current)

        self._transport_mock.write.assert_not_called()

    def test_set_voltage(self):
        """Set voltage within allowed range."""
        # Arrange
        allowed_voltages = [Tenma72_13350.MAX_VOLTAGE, 0]
        channels = [None, 2]
        expected_calls = [
            call.write(f"VSET:{allowed_voltages[0]:.3f}\n".encode("ascii")),
            call.write(f"VSET{channels[1]}:{allowed_voltages[1]:.3f}\n".encode("ascii"))
        ]
        # Act
        for e, voltage in enumerate(allowed_voltages):
            self.psu.set_voltage(voltage, channels[e])

        self._transport_mock.write.assert_has_calls(expected_calls)

    def test_set_voltage_excepts(self):
        """See that values out-of-range cause an exception."""
        # Arrange
        disallowed_voltages = [Tenma72_13350.MAX_VOLTAGE + 0.1, -0.1]
        # Act
        for voltage in disallowed_voltages:
            with self.assertRaises(ValueError):
                self.psu.set_voltage(voltage)

        self._transport_mock.write.assert_not_called()

    def test_enable_disable_output(self):
        """Test enabling and disabling output. At implemented class this should work."""
        # Arrange
        enable_cmd = "OUT:1\n"
        disable_cmd = "OUT:0\n"
        self._transport_mock.read.side_effect = [b"\x02\n", b"\x00\n"]
        expected_enable_calls = [
            call(enable_cmd.encode("ascii")),
            call(b"STATUS?\n")
        ]
        expected_disable_calls = [
            call(disable_cmd.encode("ascii")),
            call(b"STATUS?\n")
        ]
        # Act
        self.psu.enable_output(True)

        # Assert
        self._transport_mock.write.assert_has_calls(expected_enable_calls)
        self._transport_mock.write.reset_mock()

        # Act
        self.psu.enable_output(False)

        # Assert
        self._transport_mock.write.assert_has_calls(expected_disable_calls)

    def test_get_dhcp(self):
        """Get the DHCP state."""
        # Arrange
        cmd = ":SYST:DHCP?\n"
        expected_dhcp = 1
        self._transport_mock.read_until_timeout.return_value = f"{expected_dhcp}".encode("ascii")
        # Act
        state = self.psu.get_dhcp()
        # Assert
        self.assertEqual(expected_dhcp, state)
        self._transport_mock.write.assert_called_once_with(bytes(cmd, "ascii"))

    def test_set_dhcp(self):
        """Set DHCP on/off."""
        # Arrange
        dhcp = 0
        expected_calls = [
            call.write(f":SYST:DHCP {dhcp}\n".encode("ascii")),
        ]
        # Act
        self.psu.set_dhcp(dhcp)

        self._transport_mock.write.assert_has_calls(expected_calls)

    def test_get_ip_address(self):
        """Get the IP address."""
        # Arrange
        cmd = ":SYST:IPAD?\n"
        expected_ipad = "192.168.1.90"
        self._transport_mock.read_until_timeout.return_value = f"{expected_ipad}".encode("ascii")
        # Act
        ipad = self.psu.get_ip_address()
        # Assert
        self.assertEqual(expected_ipad, ipad)
        self._transport_mock.write.assert_called_once_with(bytes(cmd, "ascii"))

    def test_set_ip_address(self):
        """Set an IP address."""
        # Arrange
        ip_address = "192.168.1.90"
        expected_calls = [
            call.write(f":SYST:IPAD {ip_address}\n".encode("ascii")),
        ]
        # Act
        self.psu.set_ip_address(ip_address)

        self._transport_mock.write.assert_has_calls(expected_calls)

    def test_get_subnet_mask(self):
        """Get the subnet mask."""
        # Arrange
        cmd = ":SYST:SMASK?\n"
        expected_ipad = "255.255.255.0"
        self._transport_mock.read_until_timeout.return_value = f"{expected_ipad}".encode("ascii")
        # Act
        ipad = self.psu.get_subnet_mask()
        # Assert
        self.assertEqual(expected_ipad, ipad)
        self._transport_mock.write.assert_called_once_with(bytes(cmd, "ascii"))

    def test_set_subnet_mask(self):
        """Set a subnet mask."""
        # Arrange
        subnet_mask = "255.255.255.0"
        expected_calls = [
            call.write(f":SYST:SMASK {subnet_mask}\n".encode("ascii")),
        ]
        # Act
        self.psu.set_subnet_mask(subnet_mask)

        self._transport_mock.write.assert_has_calls(expected_calls)

    def test_get_gateway_address(self):
        """Get the gateway address."""
        # Arrange
        cmd = ":SYST:GATE?\n"
        expected_gateway = "192.168.0.1"
        self._transport_mock.read_until_timeout.return_value = f"{expected_gateway}".encode("ascii")
        # Act
        gateway = self.psu.get_gateway_address()
        # Assert
        self.assertEqual(expected_gateway, gateway)
        self._transport_mock.write.assert_called_once_with(bytes(cmd, "ascii"))

    def test_set_gateway_address(self):
        """Set a gateway address."""
        # Arrange
        gateway_address = "192.168.0.1"
        expected_calls = [
            call.write(f":SYST:GATE {gateway_address}\n".encode("ascii")),
        ]
        # Act
        self.psu.set_gateway_address(gateway_address)

        self._transport_mock.write.assert_has_calls(expected_calls)

    def test_get_ip_port(self):
        """Get the IP port."""
        # Arrange
        cmd = ":SYST:PORT?\n"
        expected_port = 5990
        self._transport_mock.read_until_timeout.return_value = f"{expected_port}".encode("ascii")
        # Act
        port = self.psu.get_ip_port()
        # Assert
        self.assertEqual(expected_port, port)
        self._transport_mock.write.assert_called_once_with(bytes(cmd, "ascii"))

    def test_set_ip_port(self):
        """Set an IP port."""
        # Arrange
        port = 5990
        expected_calls = [
            call.write(f":SYST:PORT {port}\n".encode("ascii")),
        ]
        # Act
        self.psu.set_ip_port(port)

        self._transport_mock.write.assert_has_calls(expected_calls)


if __name__ == '__main__':
    unittest.main()



"""
Aviosys, relay unit device.

The qmi.instruments.aviosys package provides support for:
- IP Power 9850 relay unit.
"""
# Alternative, QMI naming convention approved name
from qmi.instruments.aviosys.ippower import IPPower9850 as Aviosys_IpPower9850
from qmi.instruments.aviosys.ippower import PowerState, PowerSocket, Command



"""Instrument driver for the Polulo maestro servo controller"""

import logging
from typing import Dict, Generator, List, Optional, Tuple
from time import sleep

from qmi.core.context import QMI_Context
from qmi.core.instrument import QMI_Instrument, QMI_InstrumentIdentification
from qmi.core.rpc import rpc_method
from qmi.core.transport import create_transport
from qmi.core.exceptions import QMI_InstrumentException
import warnings

# Global variable holding the logger for this module.
_logger = logging.getLogger(__name__)


class Pololu_Maestro(QMI_Instrument):
    """Instrument driver for the Pololu Maestro servo controller."""

    # Baudrate of instrument.
    BAUDRATE = 9600

    # Instrument should respond within 2 seconds.
    RESPONSE_TIMEOUT = 2.0

    # Default number of channels.
    DEFAULT_NUM_CHANNELS = 6

    # Default minimum and maximum settable values.
    # These are determined by the bit lengths for the values (16 bits)
    DEFAULT_MIN_VALUE = 0
    DEFAULT_MAX_VALUE = 16383

    # Default minimum and maximum accleration.
    DEFAULT_MIN_ACCELERATION = 0
    DEFAULT_MAX_ACCELERATION = 255

    # Errors.
    ERRORS = {
        1: "Serial signal error: A hardware-level error that occurs when a byte’s stop bit is not detected at the\
            expected place. This can occur if you are communicating at a baud rate that differs from the Maestro’s\
            baud rate.",
        2: "Serial overrun error: A hardware-level error that occurs when the UART’s internal buffer fills up. This\
            should not occur during normal operation.",
        4: "Serial bugger full: A firmware-level error that occurs when the firmware’s buffer for bytes received on\
            the RX line is full and a byte from RX has been lost as a result. This error should not occur during normal\
            operation.",
        8: "Serial CRC error: This error occurs when the Maestro is running in CRC-enabled mode and the cyclic\
            redundancy check (CRC) byte at the end of the command packet does not match what the Maestro has\
            computed as that packet’s CRC. In such a case, the Maestro ignores the command packet and generates a\
            CRC error.",
        16: "Serial protocol error: This error occurs when the Maestro receives an incorrectly formatted or\
             nonsensical command packet. For example, if the command byte does not match a known command or an unfinished\
             command packet is interrupted by another command packet, this error occurs.",
        32: "Serial timeout: When the serial timeout is enabled, this error occurs whenever the timeout period has\
             elapsed without the Maestro receiving any valid serial commands. This timeout error can be used to make\
             the servos return to their home positions in the event that serial communication between the Maestro and its\
             controller is disrupted.",
        64: "Script stack error: This error occurs when a bug in the user script has caused the stack to overflow or\
             underflow. Any script command that modifies the stack has the potential to cause this error. The stack depth\
             is 32 on the Micro Maestro and 126 on the Mini Maestros.",
        128: "Script call stack error: This error occurs when a bug in the user script has caused the call stack to\
              overflow or underflow. An overflow can occur if there are too many levels of nested subroutines, or a subroutine\
              calls itself too many times. The call stack depth is 10 on the Micro Maestro and 126 on the Mini Maestros. An\
              underflow can occur when there is a return without a corresponding subroutine call. An underflow will occur if\
              you run a subroutine using the “Restart Script at Subroutine” serial command and the subroutine terminates with a\
              return command rather than a quit command or an infinite loop.",
        256: "Script program counter error: This error occurs when a bug in the user script has caused the program counter\
              (the address of the next instruction to be executed) to go out of bounds. This can happen if your program is not\
              terminated by a quit, return, or infinite loop."
    }

    def __init__(
            self, context: QMI_Context, name: str, transport: str, num_channels: int = DEFAULT_NUM_CHANNELS,
            channels_min_max_targets: Optional[Dict[int, Tuple[int, int]]] = None,
            channels_min_max_speeds: Optional[Dict[int, Tuple[int, int]]] = None,
            channels_min_max_accelerations: Optional[Dict[int, Tuple[int, int]]] = None) -> None:
        """Initialize driver.

        Parameters:
            context:                        The QMI context
            name:                           The name of the instrument instance
            transport:                      QMI transport descriptor to connect to the instrument.
            num_channels:                   Number of channels that can be controlled via the controller.
            channels_min_max_targets:       The minimum and maximum target values for each channel as dictionary, where the value is a tuple
                                            of the min and max.
            channels_min_max_speeds:        The minimum and maximum target values for each channel as dictionary, where the value is a tuple
                                            of the min and max.
            channels_min_max_accelerations: The minimum and maximum target values for each channel as dictionary, where the value is a tuple
                                            of the min and max.
        """
        super().__init__(context, name)
        self._transport = create_transport(transport, default_attributes={
                                           "baudrate": self.BAUDRATE})
        # This device number defaults to 0x0C (or 12 in decimal).
        device_nr = 0x0C
        # Command lead-in and device number are sent for each Pololu serial command.
        self._cmd_lead_in = chr(0xAA) + chr(device_nr)
        # Track the target value of each channel
        self._targets: List[int] = [0] * num_channels
        # Store the minimum and maximum target values of each channel
        self._min_target_values: List[int] = [
            self.DEFAULT_MIN_VALUE] * num_channels
        self._max_target_values: List[int] = [
            self.DEFAULT_MAX_VALUE] * num_channels

        channels_min_max_targets = channels_min_max_targets if channels_min_max_targets else {}
        channels_min_max_speeds = channels_min_max_speeds if channels_min_max_speeds else {}
        channels_min_max_accelerations = channels_min_max_accelerations if channels_min_max_accelerations else {}
        # Overwrite channels that have a provided min/max target value
        for c in channels_min_max_targets.keys():
            self._min_target_values[c] = channels_min_max_targets[c][0]
            self._max_target_values[c] = channels_min_max_targets[c][1]
        # Store the minimum and maximum speeds of each channel
        self._min_speeds: List[int] = [
            self.DEFAULT_MIN_VALUE] * num_channels
        self._max_speeds: List[int] = [
            self.DEFAULT_MAX_VALUE] * num_channels
        # Overwrite channels that have a provided min/max speed
        for c in channels_min_max_speeds.keys():
            self._min_speeds[c] = channels_min_max_speeds[c][0]
            self._max_speeds[c] = channels_min_max_speeds[c][1]
        # Store the minimum and maximum accelerations of each channel
        self._min_accelerations: List[int] = [
            self.DEFAULT_MIN_ACCELERATION] * num_channels
        self._max_accelerations: List[int] = [
            self.DEFAULT_MAX_ACCELERATION] * num_channels
        # Overwrite channels that have a provided min/max acceleration
        for c in channels_min_max_accelerations.keys():
            self._min_accelerations[c] = channels_min_max_accelerations[c][0]
            self._max_accelerations[c] = channels_min_max_accelerations[c][1]

        self._num_channels = num_channels

    @staticmethod
    def _compile_write_command(cmd: int, channel: int, value: int) -> str:
        """Compile the command into one command string. The input target value must be divided into
        the least significant and the most significant byte.

        Parameters:
            channel:    Channel number.
            value:      Value for the command.
        """
        lsb = value & 0x7F
        msb = (value >> 7) & 0x7F
        return chr(cmd) + chr(channel) + chr(lsb) + chr(msb)

    def _get_high_bits(self, num: int) -> Generator[int, None, None]:
        """Get the bits that are set high.

        Returns:
            Bits that are high.
        """
        while num:
            bit = num & (~num+1)
            yield bit
            num ^= bit

    def _get_errors(self) -> List[str]:
        """Get errors.

        Returns:
            List of errors.
        """
        self._transport.discard_read()  # Discard possible read buffer
        cmd = self._cmd_lead_in + chr(0x21)
        self._transport.write(bytes(cmd, "latin-1"))
        bin_err = self._transport.read(
            nbytes=2, timeout=self.RESPONSE_TIMEOUT)
        int_err = int.from_bytes(bin_err, "big")
        errors = []
        # loop over bits and log the errors
        for b in self._get_high_bits(int_err):
            errors.append(self.ERRORS[b])

        return errors

    def _write(self, cmd: str) -> None:
        """Send write command to instrument.

        Parameters:
            cmd: The command string to be written
        """
        cmd = self._cmd_lead_in + cmd
        self._transport.write(bytes(cmd, "latin-1"))
        # Sleep over the bits to be sent
        sleep(len(cmd) / self.BAUDRATE * 100)

        errs = self._get_errors()
        if errs:
            formatted_errs = '\n'.join(map(str, errs))
            raise QMI_InstrumentException(
                f"Pololu Maestro servo command {cmd} resulted in the following error(s).\n{formatted_errs}")

    def _ask(self, cmd: str) -> int:
        """Send ask command to instrument.

        Parameters:
            cmd: The command string to be written.

        Raises:
            QMI_InstrumentException if the servo command returns an error.

        Returns:
            The queried value.
        """
        cmd = self._cmd_lead_in + cmd
        self._transport.write(bytes(cmd, "latin-1"))
        # Sleep over the bits to be sent
        sleep(len(cmd) / self.BAUDRATE * 100)
        # Read back the response bits
        lsb = ord(self._transport.read(
            nbytes=1, timeout=self.RESPONSE_TIMEOUT))
        msb = ord(self._transport.read(
            nbytes=1, timeout=self.RESPONSE_TIMEOUT))

        errs = self._get_errors()
        if errs:
            formatted_errs = '\n'.join(map(str, errs))
            raise QMI_InstrumentException(
                f"Pololu Maestro servo command {cmd} resulted in the following error(s).\n{formatted_errs}")

        return (msb << 8) + lsb

    def _check_channel(self, channel) -> None:
        """Check the channel number.

        Raises:
            ValueError the controller does not support that channel.
        """
        if channel < 0 or channel >= self._num_channels:
            raise ValueError(
                f"Maestro Micro Servo Controller has channels 0 to {self._num_channels - 1}."
            )

    @rpc_method
    def open(self) -> None:
        self._check_is_closed()
        _logger.info("Opening connection to %s", self._name)
        self._transport.open()
        super().open()

    @rpc_method
    def close(self) -> None:
        self._check_is_open()
        _logger.info("Closing connection to %s", self._name)
        self._transport.close()
        super().close()

    @rpc_method
    def get_idn(self) -> QMI_InstrumentIdentification:
        """Return QMI_InstrumentIdentification instance."""
        self._check_is_open()
        return QMI_InstrumentIdentification(
            vendor="Pololu", model="Maestro Micro Servo Controller", serial=None, version=None
        )

    @rpc_method
    def set_target(self, channel: int, value: int) -> None:
        """Set channel to a specified target value. The target value is of units 0.25us. For example, if you want to
        set the target to 1500 us, you need to give as input 6000:
        (1500×4 = 6000 = 01011101110000 in binary)
        Servo will begin moving based on speed and acceleration parameters previously set.
        This method will check against the maximum and minimum target value of the channel. If the value to be set is
        out of the min/max range, the method will set the target value to the min/max.

        Parameters:
            channel:    Channel number.
            value:      The target value in units of quarter microseconds.
        """
        self._check_is_open()
        self._check_channel(channel)

        target = value
        if target < self._min_target_values[channel]:
            target = self._min_target_values[channel]
            _logger.warning(
                "Target value for channel %i below minimum. Setting to %i", channel, target)
        elif self._max_target_values[channel] < target:
            target = self._max_target_values[channel]
            _logger.warning(
                "Target value for channel %i above maximum. Setting to %i", channel, target)

        cmd = self._compile_write_command(0x04, channel, target)
        self._write(cmd)
        self._targets[channel] = target

    @rpc_method
    def get_targets(self) -> List[int]:
        """Get the current targets for each channel.

        Returns:
            The target values for each channel.
        """
        return self._targets

    @rpc_method
    def get_position(self, channel: int) -> int:
        """Get the current position of the specified channel in quarter-microseconds.

        Parameters:
            channel: The channel number.

        Returns:
            The position in quarter-microsecond units.
        """
        self._check_is_open()
        self._check_channel(channel)

        cmd = chr(0x10) + chr(channel)
        return self._ask(cmd)

    @rpc_method
    def set_speed(self, channel: int, speed: int) -> None:
        """Set speed of channel. This option specifies the speed of the servo in units of
        0.25 us / 10 ms => speed * 1 us / 40 ms.

        Parameters:
            channel:    The channel number.
            speed:      The target speed.

        Raises:
            ValueError if speed is out of allowable range.
        """
        self._check_is_open()
        self._check_channel(channel)
        min_speed = self._min_speeds[channel]
        max_speed = self._max_speeds[channel]
        if not min_speed <= speed <= max_speed:
            err_msg = f"Set speed [{speed}] not in the allowed range [{min_speed}, {max_speed}]"
            raise ValueError(err_msg)

        cmd = self._compile_write_command(0x07, channel, speed)
        self._write(cmd)

    @rpc_method
    def set_acceleration(self, channel: int, acceleration: int) -> None:
        """Set acceleration of a channel. This option specifies the acceleration of the servo
        in units of 0.25 us / 10 ms / 80 ms => acceleration * 1 us / 3.2 ms .

        Parameters:
            channel:        The channel number.
            acceleration:   The target acceleration.

        Raises:
            ValueError if acceleration not within 0 to 255
        """
        self._check_is_open()
        self._check_channel(channel)
        min_acc = self._min_accelerations[channel]
        max_acc = self._max_accelerations[channel]
        if not min_acc <= acceleration <= max_acc:
            err_msg = f"Set acceleration [{acceleration}] not in the allowed range [{min_acc}, {max_acc}]"
            raise ValueError(err_msg)

        cmd = self._compile_write_command(0x09, channel, acceleration)
        self._write(cmd)

    @rpc_method
    def go_home(self):
        """Send all servos and outputs to their home values."""
        self._check_is_open()
        self._write(chr(0x22))

    @rpc_method
    def set_min_target(self, channel: int, value: int) -> None:
        """Set the minimum target value for a channel.

        Parameters:
            channel:    The channel number.
            value:      The new minimum target value.
        """
        if value < self.DEFAULT_MIN_VALUE or value > self._max_target_values[channel]:
            raise ValueError(
                f"Minimum value invalid, should be in range [{self.DEFAULT_MIN_VALUE},\
                    {self._max_target_values[channel]}")

        self._min_target_values[channel] = value

    @rpc_method
    def get_min_target(self, channel: int) -> int:
        """Get the minimum target value for a channel.

        Parameters:
            channel:    The channel number.
        """
        return self._min_target_values[channel]

    @rpc_method
    def set_max_target(self, channel: int, value: int) -> None:
        """Set the maximum target that can be set for a channel.

        Parameters:
            channel:    The channel number.
            value:      The new maximum target value.
        """
        if value <= self._min_target_values[channel] or value > self.DEFAULT_MAX_VALUE:
            raise ValueError(
                f"Maximum value invalid, should be in range [{self._min_target_values[channel] + 1},\
                    {self.DEFAULT_MAX_VALUE}]")

        self._max_target_values[channel] = value

    @rpc_method
    def get_max_target(self, channel: int) -> int:
        """Get the maximum target that can be set for a channel.

        Parameters:
            channel:    The channel number.
        """
        return self._max_target_values[channel]

    @rpc_method
    def set_max(self, channel: int, value: int) -> None:
        """Set the maximum target that can be set for a channel.

        Parameters:
            channel:    The channel number.
            value:      The new maximum target value.
        """
        warnings.warn(
            f"{self.set_max.__name__} has been deprecated. Please use {self.set_max_target.__name__}.",
            DeprecationWarning)
        self.set_max_target(channel, value)

    @rpc_method
    def set_min(self, channel: int, value: int) -> None:
        """Set the minimum target that can be set for a channel.

        Parameters:
            channel:    The channel number.
            value:      The new maximum target value.
        """
        warnings.warn(
            f"{self.set_min.__name__} has been deprecated. Please use {self.set_min_target.__name__}.",
            DeprecationWarning)
        self.set_min_target(channel, value)

    @rpc_method
    def set_target_value(self, channel: int, value: int) -> None:
        """Set channel to a specified target value. The target value is of units 0.25us. For example, if you want to
        set the target to 1500 us, you need to give as input 6000:
        (1500×4 = 6000 = 01011101110000 in binary)
        Servo will begin moving based on speed and acceleration parameters previously set.
        This method will check against the maximum and minimum target value of the channel. If the value to be set is
        out of the min/max range, the method will set the target value to the min/max.

        Parameters:
            channel:    Channel number.
            value:      The target value in units of quarter microseconds.
        """
        warnings.warn(
            f"{self.set_target_value.__name__} has been deprecated. Please use {self.set_target.__name__}.",
            DeprecationWarning)
        self.set_target(channel, value)

    @rpc_method
    def get_value(self, channel: int) -> int:
        """Get the current position of the specified channel in quarter-microseconds.

        Parameters:
            channel: The channel number.

        Returns:
            The position in quarter-microsecond.
        """
        warnings.warn(
            f"{self.get_value.__name__} has been deprecated. Please use {self.get_position.__name__}.",
            DeprecationWarning)
        return self.get_position(channel)

    @rpc_method
    def move_up(self, channel: int) -> None:
        """Move the servo to maximal value.

        Parameters:
            channel: The channel number.
        """
        warnings.warn(
            f"{self.move_up.__name__} has been deprecated. There is no replacement for this.",
            DeprecationWarning)

    @rpc_method
    def move_down(self, channel: int) -> None:
        """Move the servo to minimal value.

        Parameters:
            channel: The channel number.
        """
        warnings.warn(
            f"{self.move_down.__name__} has been deprecated. There is no replacement for this.",
            DeprecationWarning)



"""QMI driver for the Digilent Analog Discovery 2 board."""

import logging
import time
from enum import Enum

import numpy as np
import pydwf
import pydwf.core.dwf_device
import pydwf.utilities
from pydwf import DwfLibrary

from qmi.core.context import QMI_Context
from qmi.core.exceptions import QMI_TimeoutException, QMI_InstrumentException, QMI_UsageException
from qmi.core.instrument import QMI_Instrument
from qmi.core.rpc import rpc_method

# Global variable holding the logger for this module.
_logger = logging.getLogger(__name__)


class OnClose(Enum):
    """On close behaviour of the device."""
    CONTINUE = 0
    STOP = 1
    SHUTDOWN = 2


class Filter(Enum):
    """Analog acquisition filter"""

    DECIMATE = 0  # Store every Nth ADC conversion, where N = ADC frequency /acquisition frequency.
    AVERAGE = 1  # Store the average of N ADC conversions.
    MINMAX = 2  # Store interleaved, the minimum and maximum values, of 2xN conversions.


class AnalogDiscovery2(QMI_Instrument):
    """QMI driver for the Analog Discovery 2 board."""

    def __init__(self, context: QMI_Context, name: str, serial_number: str) -> None:
        super().__init__(context, name)
        self._serial_number = serial_number
        self._device: pydwf.core.dwf_device.DwfDevice | None = None

    @rpc_method
    def open(self) -> None:
        _logger.info("Opening Analog Discovery 2 with serial nr %s ...", self._serial_number)
        self._check_is_closed()
        dwf = self._load_library()
        self._device = self._open_device(dwf)
        super().open()

    def _open_device(self, dwf: DwfLibrary) -> pydwf.core.dwf_device.DwfDevice:
        """Try to open the device for control.

        Returns:
            DwfDevice:               An opened DWF device instance.

        Raises:
            QMI_InstrumentException: If an error was reported by the 'openDwfDevice' DWF C library function.
            ValueError:              Any error in pydwf caused by the underlying C API or otherwise.
        """
        try:
            return pydwf.utilities.openDwfDevice(dwf, serial_number_filter=self._serial_number)
        except pydwf.DwfLibraryError as library_error:
            raise QMI_InstrumentException(str(library_error)) from library_error
        except pydwf.PyDwfError as library_error:
            raise ValueError(str(library_error)) from library_error

    @staticmethod
    def _load_library() -> DwfLibrary:
        try:
            return pydwf.DwfLibrary()
        except pydwf.PyDwfError as library_error:
            raise QMI_UsageException('Could not open library. '
                                     'Are you sure that pydwf is correctly installed?') from library_error

    @rpc_method
    def close(self) -> None:
        _logger.info("Closing Analog Discovery 2 with serial nr %s ...", self._serial_number)
        self._check_is_open()
        assert self._device is not None
        try:
            self._device.close()
        except pydwf.DwfLibraryError as library_error:
            raise QMI_InstrumentException(str(library_error)) from library_error
        self._device = None
        super().close()

    @rpc_method
    def prepare_analog_output_channel_for_static_output(self, channel: int, voltage: float) -> None:
        """Prepare analog output channel for static output.

        Parameters:
            channel: Channel to prepare for static output.
            voltage: Initial voltage.

        Raises:
            QMI_UsageException:      Raised when a device is used before it is opened.
            QMI_InstrumentException: Raised when an unexpected instrument error occurs.
        """
        if self._device is None:
            raise QMI_UsageException("Device should be opened before use.")

        analog_out = self._device.analogOut
        try:
            analog_out.nodeFunctionSet(channel, pydwf.DwfAnalogOutNode.Carrier, pydwf.DwfAnalogOutFunction.Square)
            analog_out.idleSet(channel, pydwf.DwfAnalogOutIdle.Initial)
            analog_out.nodeAmplitudeSet(channel, pydwf.DwfAnalogOutNode.Carrier, voltage)
            analog_out.nodeEnableSet(channel, pydwf.DwfAnalogOutNode.Carrier, True)
        except pydwf.DwfLibraryError as library_error:
            raise QMI_InstrumentException(str(library_error)) from library_error

    @rpc_method
    def set_analog_output_voltage_output(self, channel: int, voltage: float) -> None:
        """Set analog output channel to a target voltage.

        Parameters:
            channel: Target channel.
            voltage: Target voltage level.

        Raises:
            QMI_UsageException:      Raised when a device is used before it is opened.
            QMI_InstrumentException: Raised when an unexpected instrument error occurs.
        """
        if self._device is None:
            raise QMI_UsageException("Device should be opened before use.")

        try:
            self._device.analogOut.nodeAmplitudeSet(channel, pydwf.DwfAnalogOutNode.Carrier, voltage)
        except pydwf.DwfLibraryError as library_error:
            raise QMI_InstrumentException(str(library_error)) from library_error

    @rpc_method
    def get_analog_output_voltage_output(self, channel: int) -> float:
        """Get the set voltage level for the analog output.

        Parameters:
            channel:          The target channel to query.

        Returns:
            voltage_setpoint: The set voltage level of the specified channel.

        Raises:
            QMI_UsageException:      Raised when a device is used before it is opened.
            QMI_InstrumentException: Raised when an unexpected instrument error occurs.
        """
        if self._device is None:
            raise QMI_UsageException("Device should be opened before use.")

        try:
            voltage_setpoint = self._device.analogOut.nodeAmplitudeGet(channel, pydwf.DwfAnalogOutNode.Carrier)
        except pydwf.DwfLibraryError as library_error:
            raise QMI_InstrumentException(str(library_error)) from library_error

        return voltage_setpoint

    @rpc_method
    def set_device_on_close(self, on_close: OnClose) -> None:
        """Set on close behaviour.

        Parameters:
            on_close: OnClose enum: either CONTINUE, STOP or SHUTDOWN after closing the device.

        Raises:
            QMI_UsageException:      Raised when a device is used before it is opened.
            QMI_InstrumentException: Raised when an unexpected instrument error occurs.
        """
        if self._device is None:
            raise QMI_UsageException("Device should be opened before use.")

        try:
            self._device.paramSet(pydwf.DwfDeviceParameter.OnClose, on_close.value)
        except pydwf.DwfLibraryError as library_error:
            raise QMI_InstrumentException(str(library_error)) from library_error

    @rpc_method
    def reset_analog_input(self):
        """Resets and configures analog input parameters to default values.

        Raises:
            QMI_UsageException:      Raised when a device is used before it is opened.
            QMI_InstrumentException: Raised when an unexpected instrument error occurs.
        """
        if self._device is None:
            raise QMI_UsageException("Device should be opened before use.")

        analog_in = self._device.analogIn
        try:
            analog_in.reset()
        except pydwf.DwfLibraryError as library_error:
            raise QMI_InstrumentException(str(library_error)) from library_error

    @rpc_method
    def prepare_analog_input_sample(
        self,
        channel: int,
        voltage_range: float,
        acquisition_frequency: float,
        adc_filter: Filter = Filter.AVERAGE,
        voltage_offset: float = 0.0,
        buffer_size: int | None = None
    ) -> None:
        """Prepare readout voltage on specified channel.

        Parameters:
            channel:               The channel to prepare.
            voltage_offset:        Offset in volts.
            voltage_range:         Range in voltage defined as peak to peak value centered around the offset value.
            acquisition_frequency: Acquisition frequency in Hz. The device samples ADC operates at 100 MHz.
                                   If an acquisition frequency lower than 100 MHz is provided. The samples will be
                                   filtered using the provided filter rule.
                                   Note that the acquisition is set for both channels.
            adc_filter:            Filter rule to apply to ADC samples. See analog_discovery.Filter for documentation.
            buffer_size:           Adjust the buffer size of the analog in instrument: 16 <= buffer_size <= 8192.

        Raises:
            QMI_UsageException:      Raised when a device is used before it is opened.
            QMI_InstrumentException: Raised when an unexpected instrument error occurs.
        """
        if self._device is None:
            raise QMI_UsageException("Device should be opened before use.")

        analog_in = self._device.analogIn
        try:
            analog_in.channelEnableSet(channel, True)
            analog_in.frequencySet(acquisition_frequency)
            if buffer_size is not None:
                analog_in.bufferSizeSet(buffer_size)

            analog_in.channelOffsetSet(channel, channel_offset=voltage_offset)
            analog_in.channelRangeSet(channel, channel_range=voltage_range)
            analog_in.channelFilterSet(channel, pydwf.DwfAnalogInFilter(adc_filter.value))
            analog_in.configure(False, False)
        except pydwf.DwfLibraryError as library_error:
            raise QMI_InstrumentException(str(library_error)) from library_error

        # Wait at least 2 seconds with Analog Discovery for the offset to stabilize, before the first reading after
        # device open or offset/range change
        time.sleep(2)

    @rpc_method
    def get_analog_input_sample(self, channel: int) -> float:
        """Readout voltage on specified channel.

        Parameters:
            channel: the target channel to query.

        Raises:
            QMI_UsageException: Raised when a device is used before it is opened.
            QMI_InstrumentException: QMI_InstrumentException is raised when an unexpected instrument error occurs.
        """
        if self._device is None:
            raise QMI_UsageException("Device should be opened before use.")

        analog_in = self._device.analogIn
        try:
            analog_in.status(False)
            sample = analog_in.statusSample(channel)
        except pydwf.DwfLibraryError as library_error:
            raise QMI_InstrumentException(str(library_error)) from library_error

        return sample

    @rpc_method
    def get_analog_input_acquire_samples(self, channel: int, timeout: float | None = None) -> np.ndarray:
        """Acquire analog input samples.

        Parameters:
            channel: The target channel to acquire samples from.
            timeout: Time before timeout exception is raised while waiting on sampling acquisition to finish. This
                     method blocks indefinitely if 'None' is provided and the measurement is not finished.

        Raises:
            QMI_UsageException:      Raised when a device is used before it is opened.
            QMI_TimeoutException:    Raised when a timeout occurs while waiting for the sampling acquisition to finish.
            QMI_InstrumentException: Raised when an unexpected instrument error occurs.
        """
        if self._device is None:
            raise QMI_UsageException("Device should be opened before use.")

        analog_in = self._device.analogIn
        try:
            # Reconfigure the instrument (reset the auto trigger) and start the acquisition.
            analog_in.configure(True, True)
            time_start = time.monotonic()
            while True:
                status = analog_in.status(True)
                if status == pydwf.DwfState.Done:
                    break
                if (timeout is not None) and (time.monotonic() - time_start) > timeout:
                    raise QMI_TimeoutException('Timeout while waiting for samples.')
                time.sleep(0.01)
            buffer_size = analog_in.bufferSizeGet()
            samples = analog_in.statusData(channel, buffer_size)
        except pydwf.DwfLibraryError as library_error:
            raise QMI_InstrumentException(str(library_error)) from library_error

        return samples

    @rpc_method
    def prepare_analog_input_record(
        self,
        channel: int,
        voltage_range: float,
        record_length: float,
        acquisition_frequency: float,
        voltage_offset: float = 0.0,
        adc_filter: Filter = Filter.AVERAGE
    ) -> None:
        """Prepare an analog input record for specified channel.

        Raises:
            QMI_UsageException: Raised when a device is used before it is opened.
            QMI_InstrumentException: QMI_InstrumentException is raised when an unexpected instrument error occurs.
        """
        if self._device is None:
            raise QMI_UsageException("Device should be opened before use.")

        analog_in = self._device.analogIn
        try:
            analog_in.channelEnableSet(channel, True)
            analog_in.channelOffsetSet(channel, channel_offset=voltage_offset)
            analog_in.channelRangeSet(channel, channel_range=voltage_range)
            analog_in.acquisitionModeSet(pydwf.DwfAcquisitionMode.Record)
            analog_in.frequencySet(acquisition_frequency)
            analog_in.recordLengthSet(record_length)
            analog_in.channelFilterSet(channel, pydwf.DwfAnalogInFilter(adc_filter.value))
        except pydwf.DwfLibraryError as library_error:
            raise QMI_InstrumentException(str(library_error)) from library_error

        # Wait at least 2 seconds with Analog Discovery for the offset to stabilize, before the first reading after
        # device open or offset/range change
        time.sleep(2)

    @rpc_method
    def get_analog_input_record(self, channel: int, amount_samples: int | None = None) -> list[float]:
        """Obtain a record of input voltages from specified channel.

        Parameters:
            channel:        The target channel to query.
            amount_samples: The amount of samples to take. Only relevant when record length is set to -1.

        Returns:
            sample_record:  A record of input voltages.

        Raises:
            QMI_UsageException:      Raised when a device is used before it is opened.
            QMI_InstrumentException: Raised when data is corrupt or lost or when an unexpected instrument error occurs.
        """
        if self._device is None:
            raise QMI_UsageException("Device should be opened before use.")

        analog_in = self._device.analogIn
        sample_record: np.typing.NDArray[np.float64] = np.array([])
        try:
            if amount_samples is None:
                amount_samples_rational = analog_in.recordLengthGet() // analog_in.frequencyGet()
                amount_samples = int(amount_samples_rational)

            start_time = time.monotonic()
            analog_in.configure(False, True)
            while len(sample_record) < amount_samples and time.monotonic() < start_time + 0.5:
                status = analog_in.status(True)
                if len(sample_record) == 0 and status in (
                    pydwf.DwfState.Config,
                    pydwf.DwfState.Prefill,
                    pydwf.DwfState.Armed
                ):
                    continue

                data_available, data_lost, data_corrupt = analog_in.statusRecord()
                if data_lost > 0 or data_corrupt > 0:
                    raise QMI_InstrumentException('Data was corrupt or lost when obtaining analog input record.')
                if data_available == 0:
                    continue
                if data_available + len(sample_record) > amount_samples:
                    data_available = amount_samples - len(sample_record)

                samples = analog_in.statusData(channel, data_available)
                sample_record = np.append(sample_record, samples)

        except pydwf.DwfLibraryError as library_error:
            raise QMI_InstrumentException(str(library_error)) from library_error

        return list(sample_record)



"""Unit tests for Edwards Turbo Instrument Controller."""
import logging
from typing import cast
import unittest
from unittest.mock import Mock, patch

import qmi
from qmi.core.exceptions import QMI_InstrumentException
from qmi.core.transport import QMI_TcpTransport
from qmi.instruments.edwards import EdwardsVacuum_TIC, EdwardsVacuum_TIC_AlertId,\
    EdwardsVacuum_TIC_GaugeState, EdwardsVacuum_TIC_Priority,\
    EdwardsVacuum_TIC_PumpState, EdwardsVacuum_TIC_State


class TestTurboInstrumentControllerOpenClose(unittest.TestCase):
    """
    Open close test for Edwards TIC.
    """

    def setUp(self):
        qmi.start("test-tic-context")
        # Add patches
        patcher = patch('qmi.instruments.edwards.turbo_instrument_controller.create_transport', spec=QMI_TcpTransport)
        self._transport_mock: Mock = patcher.start().return_value
        self.addCleanup(patcher.stop)
        patcher = patch('qmi.instruments.edwards.turbo_instrument_controller.ScpiProtocol', autospec=True)
        self._scpi_mock: Mock = patcher.start().return_value
        self.addCleanup(patcher.stop)
        # Make DUTs
        self.instr: EdwardsVacuum_TIC = qmi.make_instrument(
            "test_instr", EdwardsVacuum_TIC, "transport_str")
        self.instr = cast(EdwardsVacuum_TIC, self.instr)

    def tearDown(self):
        qmi.stop()
        logging.getLogger("qmi.core.instrument").setLevel(logging.NOTSET)

    def test_open_close_opens_and_closes(self):
        """Open and close instrument."""
        self.instr.open()
        self._transport_mock.open.assert_called_once_with()
        self.instr.close()
        self._transport_mock.close.assert_called_once_with()


class TestTurboInstrumentController(unittest.TestCase):
    """
    Unit tests for Edwards TIC
    """

    def setUp(self):
        qmi.start("test-tic-context")
        # Add patches
        patcher = patch('qmi.instruments.edwards.turbo_instrument_controller.create_transport', spec=QMI_TcpTransport)
        self._transport_mock: Mock = patcher.start().return_value
        self.addCleanup(patcher.stop)
        patcher = patch('qmi.instruments.edwards.turbo_instrument_controller.ScpiProtocol', autospec=True)
        self._scpi_mock: Mock = patcher.start().return_value
        self.addCleanup(patcher.stop)
        # Make DUTs
        self.instr: EdwardsVacuum_TIC = qmi.make_instrument(
            "test_instr", EdwardsVacuum_TIC, "transport_str")
        self.instr = cast(EdwardsVacuum_TIC, self.instr)
        self.instr.open()

    def tearDown(self):
        self.instr.close()
        qmi.stop()

    def test_get_idn_gets_idn(self):
        """Get idn of device."""
        # Arrange
        actual_model = "model1"
        actual_version = "version2"
        actual_serial = "serial3"

        cmd = "?S902"

        self._scpi_mock.ask.return_value = f"=S902 {actual_model};{actual_version};{actual_serial}"
        # Act
        instr_info = self.instr.get_idn()

        # Assert
        self._scpi_mock.ask.assert_called_once_with(cmd)
        self.assertEqual(instr_info.vendor, "Edwards")
        self.assertEqual(instr_info.model, actual_model)
        self.assertEqual(instr_info.version, actual_version)
        self.assertEqual(instr_info.serial, actual_serial)

    def test_get_response_with_invalid_responses_throws_instrument_exception(self):
        """Get response with invalid reponse, throws an instrument exception."""
        # Arrange
        cmd = "?S902"

        self._scpi_mock.ask.return_value = "=A902 0;0;0"

        # Act
        # Assert
        with self.assertRaises(QMI_InstrumentException):
            self.instr.get_idn()
            self._scpi_mock.ask.assert_called_once_with(cmd)

    def test_get_tic_status_gets_status(self):
        """Get the status of the instrument."""
        # Arrange
        actual_turbo_pump_state = EdwardsVacuum_TIC_PumpState.STOPPED
        actual_backing_pump_state = EdwardsVacuum_TIC_PumpState.ACCELERATING
        actual_gauge1_state = EdwardsVacuum_TIC_GaugeState.ZEROING
        actual_relay1_state = EdwardsVacuum_TIC_State.ON
        actual_alert = EdwardsVacuum_TIC_AlertId.FILAMENT_FAIL
        actual_priority = EdwardsVacuum_TIC_Priority.WARNING

        cmd = "?V902"

        self._scpi_mock.ask.return_value = "=V902 0;5;9;0;0;4;2;3;15;1"

        # Act
        status = self.instr.get_tic_status()

        # Assert
        self._scpi_mock.ask.assert_called_once_with(cmd)
        self.assertEqual(status.turbo_pump_state, actual_turbo_pump_state)
        self.assertEqual(status.backing_pump_state, actual_backing_pump_state)
        self.assertEqual(status.gauge_1_state, actual_gauge1_state)
        self.assertEqual(status.relay_1_state, actual_relay1_state)
        self.assertEqual(status.alert, actual_alert)
        self.assertEqual(status.priority, actual_priority)

    def test_get_turbo_pump_state_gets_state(self):
        """Get state of turbo pump."""
        # Arrange
        actual_turbo_pump_state = EdwardsVacuum_TIC_PumpState.STOPPING_NORMAL_DELAY
        actual_alert = EdwardsVacuum_TIC_AlertId.NEW_ID
        actual_priority = EdwardsVacuum_TIC_Priority.OK

        cmd = "?V904"

        self._scpi_mock.ask.return_value = "=V904 3;9;0"

        # Act
        status = self.instr.get_turbo_pump_state()

        # Assert
        self._scpi_mock.ask.assert_called_once_with(cmd)
        self.assertEqual(status.state, actual_turbo_pump_state)
        self.assertEqual(status.alert, actual_alert)
        self.assertEqual(status.priority, actual_priority)

    def test_get_backing_pump_state_gets_state(self):
        """Get state of backing pump."""
        # Arrange
        actual_turbo_pump_state = EdwardsVacuum_TIC_PumpState.FAULT_BRAKING
        actual_alert = EdwardsVacuum_TIC_AlertId.OVER_PRESSURE
        actual_priority = EdwardsVacuum_TIC_Priority.OK

        cmd = "?V910"

        self._scpi_mock.ask.return_value = "=V910 6;23;0"

        # Act
        status = self.instr.get_backing_pump_state()

        # Assert
        self._scpi_mock.ask.assert_called_once_with(cmd)
        self.assertEqual(status.state, actual_turbo_pump_state)
        self.assertEqual(status.alert, actual_alert)
        self.assertEqual(status.priority, actual_priority)

    def test_turn_on_turbo_pump_turns_on_pump(self):
        """Turn on turbo pump."""
        # Arrange
        cmd = "!C904 1"

        self._scpi_mock.ask.return_value = "*C904 1"

        # Act
        self.instr.turn_on_turbo_pump()

        # Assert
        self._scpi_mock.ask.assert_called_once_with(cmd)

    def test_turn_on_turbo_pump_with_error_throws_instrument_exception(self):
        """Turn on turbo pump with error, throws instrument exception."""
        # Arrange
        cmd = "!C904 1"

        self._scpi_mock.ask.return_value = "*C904 5"

        # Act
        # Assert
        with self.assertRaises(QMI_InstrumentException):
            self.instr.turn_on_turbo_pump()
            self._scpi_mock.ask.assert_called_once_with(cmd)

    def test_turn_off_turbo_pump_turns_off_pump(self):
        """Turn off turbo pump."""
        # Arrange
        cmd = "!C904 0"

        self._scpi_mock.ask.return_value = "*C904 0"

        # Act
        self.instr.turn_off_turbo_pump()

        # Assert
        self._scpi_mock.ask.assert_called_once_with(cmd)

    def test_turn_on_backing_pump_turns_on_pump(self):
        """Turn on backing pump."""
        # Arrange
        cmd = "!C910 1"

        self._scpi_mock.ask.return_value = "*C910 1"

        # Act
        self.instr.turn_on_backing_pump()

        # Assert
        self._scpi_mock.ask.assert_called_once_with(cmd)

    def test_turn_off_backing_pump_turns_off_pump(self):
        """Turn off backing pump."""
        # Arrange
        cmd = "!C910 0"

        self._scpi_mock.ask.return_value = "*C910 0"

        # Act
        self.instr.turn_off_backing_pump()

        # Assert
        self._scpi_mock.ask.assert_called_once_with(cmd)

    def test_get_turbo_pump_speed_gets_speed(self):
        """Get the speed of the turbo pump."""
        # Arrange
        actual_turbo_pump_speed = 55.9
        actual_alert = EdwardsVacuum_TIC_AlertId.ADC_FAULT
        actual_priority = EdwardsVacuum_TIC_Priority.OK

        cmd = "?V905"

        self._scpi_mock.ask.return_value = f"=V905 {actual_turbo_pump_speed};1;0"

        # Act
        speed = self.instr.get_turbo_pump_speed()

        # Assert
        self._scpi_mock.ask.assert_called_once_with(cmd)

        self.assertEqual(speed.speed, actual_turbo_pump_speed)
        self.assertEqual(speed.alert, actual_alert)
        self.assertEqual(speed.priority, actual_priority)

    def test_get_backing_pump_speed_gets_speed(self):
        """Get the speed of the backing pump."""
        # Arrange
        actual_backing_pump_speed = 15.2
        actual_alert = EdwardsVacuum_TIC_AlertId.DX_FAULT
        actual_priority = EdwardsVacuum_TIC_Priority.OK

        cmd = "?V911"

        self._scpi_mock.ask.return_value = f"=V911 {actual_backing_pump_speed};32;0"

        # Act
        speed = self.instr.get_backing_pump_speed()

        # Assert
        self._scpi_mock.ask.assert_called_once_with(cmd)

        self.assertEqual(speed.speed, actual_backing_pump_speed)
        self.assertEqual(speed.alert, actual_alert)
        self.assertEqual(speed.priority, actual_priority)

    def test_get_turbo_pump_power_gets_power(self):
        """Get power of turbo pump."""
        # Arrange
        actual_turbo_pump_power = 10
        actual_alert = EdwardsVacuum_TIC_AlertId.ADC_FAULT
        actual_priority = EdwardsVacuum_TIC_Priority.OK

        cmd = "?V906"

        self._scpi_mock.ask.return_value = f"=V906 {actual_turbo_pump_power};1;0"

        # Act
        speed = self.instr.get_turbo_pump_power()

        # Assert
        self._scpi_mock.ask.assert_called_once_with(cmd)

        self.assertEqual(speed.power, actual_turbo_pump_power)
        self.assertEqual(speed.alert, actual_alert)
        self.assertEqual(speed.priority, actual_priority)

    def test_get_backing_pump_power_gets_power(self):
        """Get power of backing pump."""
        # Arrange
        actual_backing_pump_power = 20
        actual_alert = EdwardsVacuum_TIC_AlertId.ADC_FAULT
        actual_priority = EdwardsVacuum_TIC_Priority.OK

        cmd = "?V912"

        self._scpi_mock.ask.return_value = f"=V912 {actual_backing_pump_power};1;0"

        # Act
        speed = self.instr.get_backing_pump_power()

        # Assert
        self._scpi_mock.ask.assert_called_once_with(cmd)

        self.assertEqual(speed.power, actual_backing_pump_power)
        self.assertEqual(speed.alert, actual_alert)
        self.assertEqual(speed.priority, actual_priority)

    def test_get_pressure_gauge_1_gets_pressure(self):
        """Get pressure of gauge 1."""
        # Arrange
        actual_pressure = 5
        actual_unit = "V"
        actual_gauge_state = EdwardsVacuum_TIC_GaugeState.STRIKING
        actual_alert = EdwardsVacuum_TIC_AlertId.ADC_FAULT
        actual_priority = EdwardsVacuum_TIC_Priority.OK

        cmd = "?V913"

        self._scpi_mock.ask.return_value = f"=V913 {actual_pressure};66;6;1;0"

        # Act
        pressure = self.instr.get_pressure(1)

        # Assert
        self._scpi_mock.ask.assert_called_once_with(cmd)

        self.assertEqual(pressure.pressure, actual_pressure)
        self.assertEqual(pressure.unit, actual_unit)
        self.assertEqual(pressure.state, actual_gauge_state)
        self.assertEqual(pressure.alert, actual_alert)
        self.assertEqual(pressure.priority, actual_priority)

    def test_get_relay_state_3_gets_state(self):
        """Get state of 3."""
        # Arrange
        actual_state = EdwardsVacuum_TIC_State.OFF_GOING_ON
        actual_alert = EdwardsVacuum_TIC_AlertId.ADC_FAULT
        actual_priority = EdwardsVacuum_TIC_Priority.OK

        cmd = "?V918"

        self._scpi_mock.ask.return_value = "=V918 1;1;0"

        # Act
        state = self.instr.get_relay_state(3)

        # Assert
        self._scpi_mock.ask.assert_called_once_with(cmd)

        self.assertEqual(state.state, actual_state)
        self.assertEqual(state.alert, actual_alert)
        self.assertEqual(state.priority, actual_priority)

    def test_turn_on_relay_1_turns_on_relay(self):
        """Turn on relay 1."""
        # Arrange
        cmd = "!C917 1"

        self._scpi_mock.ask.return_value = "*C917 1"

        # Act
        self.instr.turn_on_relay(2)

        # Assert
        self._scpi_mock.ask.assert_called_once_with(cmd)

    def test_turn_off_relay_1_turns_off_relay(self):
        """Turn off relay 1."""
        # Arrange
        cmd = "!C916 0"

        self._scpi_mock.ask.return_value = "*C916 0"

        # Act
        self.instr.turn_off_relay(1)

        # Assert
        self._scpi_mock.ask.assert_called_once_with(cmd)


if __name__ == '__main__':
    unittest.main()



"""
Instrument driver for the Adwin Pro II. Inherits from Adwin_Base class in adwin.py
"""
from qmi.core.context import QMI_Context
from qmi.instruments.adwin.adwin import Adwin_Base


class Adwin_ProII(Adwin_Base):
    """Instrument driver for Adwin real-time microcontroller.

    This driver is specific for the Adwin Pro 2.
    """

    def __init__(self, context: QMI_Context, name: str, device_no: int) -> None:
        """Initialize the Adwin driver.

        Parameters:
            context: QMI context.
            name: Name for this instrument instance.
            device_no: Adwin device number.
        """
        super().__init__(context, name, device_no)



"""Instrument driver for the OZ Optics 100MC attenuator.
"""

import logging
from typing import Optional, List, NamedTuple

from qmi.core.context import QMI_Context
from qmi.core.exceptions import QMI_InstrumentException
from qmi.core.instrument import QMI_Instrument
from qmi.core.rpc import rpc_method
from qmi.core.transport import create_transport

# Global variable holding the logger for this module.
_logger = logging.getLogger(__name__)


class OZO_AttenuatorPosition(NamedTuple):
    """Representation of attenuator setpoint.

    Attributes:
        steps:       Piezo position.
        attenuation: Realized attenuation in decibels.
    """
    steps: int
    attenuation: float


class OZOptics_DD100MC(QMI_Instrument):
    """Instrument driver for the OzOptics DD-100-MC attenuator.

    NOTE: the device expects commands ending with a Carriage Return character ('\r') and responds
    with lines ending in CR/LF ('\r\n').
    """
    # Instrument should respond within 5 seconds.
    RESPONSE_TIMEOUT = 5.0
    RESPONSE_TIMEOUT_HOME_COMMAND = 30.0  # HOME command may take longer.

    def __init__(self, context: QMI_Context, name: str, transport: str) -> None:
        super().__init__(context, name)
        self._transport = create_transport(transport, default_attributes={"baudrate": 9600})

    @rpc_method
    def open(self) -> None:
        _logger.info("Opening connection to %s", self._name)
        self._transport.open()
        self._transport.discard_read()
        super().open()

    @rpc_method
    def close(self) -> None:
        _logger.info("Closing connection to %s", self._name)
        super().close()
        self._transport.close()

    def _read_response(self, timeout: Optional[float]) -> List[str]:
        """Reads a response from instrument. The response consists of multiple 'lines' ending
        with CR\LF, where the last two 'lines' are "Done\r\n" and an empty line "\r\n".

        Returns:
            response: A list of ASCII strings with the last two 'lines' removed.
        """
        self._check_is_open()
        response = []
        while True:
            line_bytes = self._transport.read_until(b"\r\n", timeout)
            line = line_bytes[:-2].decode("ascii")
            response.append(line)
            if len(response) >= 2 and response[-2] == "Done" and response[-1] == "":
                break

        response = response[:-2]  # Remove the last 2 lines.
        return response

    def _execute_command(self, command: str, timeout: Optional[float]) -> List[str]:
        """Executes a command and reads the instrument response to the command."""
        self._check_is_open()
        command_bytes = (command + "\r").encode("ascii")
        self._transport.write(command_bytes)
        return self._read_response(timeout)

    @rpc_method
    def get_configuration_display(self) -> List[str]:
        """Get display configuration string.

        Returns:
            response: A list of strings showing current display configuration.
        """
        command = "CD"
        response = self._execute_command(command, self.RESPONSE_TIMEOUT)
        return response  # Don't even try to parse the result.

    @rpc_method
    def home(self) -> None:
        """Move attenuator to home position. Using extra long timeout."""
        command = "H"
        self._execute_command(command, self.RESPONSE_TIMEOUT_HOME_COMMAND)

    @rpc_method
    def get_position(self) -> OZO_AttenuatorPosition:
        """Get current attenuator position.

        Raises:
            QMI_InstrumentException: On bad response to query command.

        Returns:
            response: Piezo position in steps and realized attenuation in decibel.
        """
        command = "D"
        response = self._execute_command(command, self.RESPONSE_TIMEOUT)

        response_ok = (len(response) == 2) and response[0].startswith("DPos:") and response[1].startswith("ATTEN:")
        if not response_ok:
            raise QMI_InstrumentException("Bad response: {!r}".format(response))

        steps = int(response[0][5:])
        attenuation = float(response[1][6:])

        return OZO_AttenuatorPosition(steps, attenuation)

    @rpc_method
    def set_attenuation(self, value: float) -> int:
        """Set new target attenuation value in Decibel.

        Parameters:
            value: The target attenuation.

        Raises:
            QMI_InstrumentException: On bad response to command.

        Returns:
            Attenuator state (0 = IDLE).
        """
        command = "A{:.2f}".format(value)
        response = self._execute_command(command, self.RESPONSE_TIMEOUT)
        response_ok = (len(response) >= 1) and response[-1].startswith("Pos:")
        if not response_ok:
            raise QMI_InstrumentException("Bad response: {!r}".format(response))

        return int(response[-1][4:])



"""Routines for data storage."""

import os
import os.path
import re
import shutil
import time
from typing import List, Optional, Any
import json
import h5py

from qmi.core.exceptions import QMI_UsageException
import qmi.data.dataset
from qmi.data.dataset import DataSet
from qmi.core.config_struct import config_struct_to_dict


def _relative_folder_path(date_str: str, time_str: str, label: str) -> str:
    """Internal function to determine the relative path to a DataFolder."""
    return os.path.join(date_str, time_str + "_" + label)


class DataFolder:
    """A DataFolder represents a collection of files from a single measurement.

    A DataFolder typically exists within a DataStore. In this case
    the DataFolder is identified by a label, date code and time code.

    The contents of a DataFolder may consist of:
      - any number of data files from the measurement;
      - a QMI configuration file and/or additional configuration files;
      - data files resulting from analysis;
      - plotted images.
    """

    def __init__(self,
                 folder_path: str,
                 label: Optional[str],
                 date_str: Optional[str],
                 time_str: Optional[str]
                 ) -> None:
        """Initialize a DataFolder instance.

        Parameters:
            folder_path: Path to the directory representing the DataFolder.
            label: Optional label of the folder (indicating type of measurement)
            date_str: Optional date code of the folder.
            time_str: Optional time code of the folder.

        Raises:
            FileNotFoundError: If the specified DataFolder does not exist.
        """

        self.folder_path = folder_path
        self.label = label
        self.date_str = date_str
        self.time_str = time_str

        if not os.path.isdir(self.folder_path):
            raise FileNotFoundError("DataFolder directory {!r} not found".format(self.folder_path))

    def __repr__(self) -> str:
        return "DataFolder({!r})".format(self.folder_path)

    def write_config(self, config: Any) -> None:
        """Write QMI configuration to a file in the data folder.
        """
        # Resolve file name for the config
        config_fn = ""
        if self.label:
            config_fn += self.label

        if self.date_str:
            config_fn += "-" + self.date_str

        if self.time_str:
            config_fn += "-" + self.time_str

        if not config_fn:
            config_fn = "config"

        full_path_fn = os.path.join(self.folder_path, config_fn + ".json")

        config_dict = config_struct_to_dict(config)

        with open(full_path_fn, "w") as output:
            json.dump(config_dict, output, indent=4, sort_keys=True)

    def copy_file(self, filename: str) -> None:
        """Copy an existing file into the data folder.

        Parameters:
            filename: Path to existing file to be copied to the data folder.

        Raises:
            FileExistsError: If the data folder already contains a file with the same name.
        """
        shutil.copy2(filename, self.folder_path)

    def write_dataset(self, ds: DataSet, file_format: str = "hdf5", overwrite: bool = False) -> None:
        """Write the specified DataSet to a new or existing file in the data folder.

        The file name will be determined from the name of the DataSet.

        Parameters:
            ds: DataSet instance to write.
            file_format: File format specification.
                "hdf5" - selects HDF5 format (default)
                "text" - selects a space-separated text format
            overwrite: Allow user to overwrite an existing dataset. Default is false.

        Raises:
            OSError: If the data folder already contains a file with the same name.
        """

        if not re.match(r"^[-_a-zA-Z0-9(),]+$", ds.name):
            raise ValueError("Invalid DataSet name {!r}".format(ds.name))

        if file_format == "hdf5":
            filename = ds.name + ".h5"
            file_path = os.path.join(self.folder_path, filename)
            with h5py.File(file_path, "w" if overwrite else "x") as f:
                qmi.data.dataset.write_dataset_to_hdf5(ds, f)

        elif file_format == "text":
            filename = ds.name + ".dat"
            file_path = os.path.join(self.folder_path, filename)
            with open(file_path, "wt" if overwrite else "xt") as f:
                qmi.data.dataset.write_dataset_to_text(ds, f)

        else:
            raise QMI_UsageException("Unknown file format {!r}".format(file_format))

    def read_dataset(self, name: str) -> DataSet:
        """Read a DataSet from the data folder.

        The file name and format will be determined from the name of
        the DataSet and the contents of the data folder.

        Parameters:
            name: Name of the dataset.

        Returns:
            DataSet instance loaded from the data folder.

        Raises:
            FileNotFoundError: If the data folder does not contain the specified dataset.
        """

        # Check that the name is safe (no path names).
        if os.path.split(name)[0]:
            raise ValueError("Invalid dataset name {!r}".format(name))

        # Look for a HDF5 file with matching name.
        file_path = os.path.join(self.folder_path, name + ".h5")
        if os.path.isfile(file_path):
            with h5py.File(file_path, "r") as f:
                if name not in f:
                    raise FileNotFoundError("No dataset {!r} found in {}".format(name, file_path))
                return qmi.data.dataset.read_dataset_from_hdf5(f[name])

        # Look for a text file with matching name.
        file_path = os.path.join(self.folder_path, name + ".dat")
        if os.path.isfile(file_path):
            with open(file_path, "rt") as f:
                return qmi.data.dataset.read_dataset_from_text(f)

        # File not found.
        raise FileNotFoundError("No dataset {!r} found in {}".format(name, self.folder_path))

    def make_hdf5file(self, name: str) -> h5py.File:
        """Create a new HDF5 file in the data folder.

        An error occurs if the specified file already exists.

        Parameters:
            name: Base name of the HDF5 file, without the extension ".h5".

        Returns:
            A `File` object representing the HDF5 file.
            See http://docs.h5py.org/ for information on how to use this object.

        Raises:
            ValueError: If the `name` has non-latin character(s).
        """
        if not re.match(r"^[-_a-zA-Z0-9(),]+$", name):
            raise ValueError("Invalid name {!r}".format(name))

        filename = name + ".h5"
        file_path = os.path.join(self.folder_path, filename)

        return h5py.File(file_path, mode="x")

    def open_hdf5file(self, name: str, write_mode: bool = False) -> h5py.File:
        """Open an existing HDF5 file in the data folder.

        Parameters:
            name: Base name of the HDF5 file, without the extension ".h5".
            write_mode: True to open the file in read/write mode, False to open the file in read-only mode.

        Returns:
            A `File` object representing the HDF5 file.
            See http://docs.h5py.org/ for information on how to use this object.

        Raises:
            ValueError: If the `name` has non-latin character(s).
        """
        if not re.match(r"^[-_a-zA-Z0-9(),]+$", name):
            raise ValueError("Invalid name {!r}".format(name))

        filename = name + ".h5"
        file_path = os.path.join(self.folder_path, filename)

        mode = "r+" if write_mode else "r"
        return h5py.File(file_path, mode=mode)


class DataStore:
    """A DataStore represents a collection of stored data.

    A DataStore instance can potentially contain data from many different
    types of measurements, performed at different times under different
    conditions.

    A DataStore instance corresponds to a folder in the file system which
    contains the actual stored files. The DataStore instance provides
    a convenient interface to store and access the data.

    The file system structure of the DataStore is as follows:
      <basedir>/<date_str>/<time_str>_<label>/<measurement_file>

    where
      <date_str> is an 8-digit string in YYYYmmdd format;
      <time_str> is a 6-digit string in HHMMSS format.

    In other words, the DataStore base directory contains a separate
    sub-directory for each date. Each of these date sub-directories contains
    a separate sub-directory for each measurement, labeled by a time code
    and label for the measurement. Each of the measurement subdirectories
    contains any number of files related to the measurement.
    """

    # True to use local time for folder names; False to use UTC.
    USE_LOCAL_TIME = True

    def __init__(self, basedir: str) -> None:
        """Initialize a DataStore instance.

        Parameters:
            basedir: Base directory for stored files.
        """
        self.basedir = basedir
        if not os.path.isdir(basedir):
            raise FileNotFoundError("DataStore base directory {!r} not found".format(basedir))

    def __repr__(self) -> str:
        return "DataStore({!r})".format(self.basedir)

    def make_folder(self,
                    label: str,
                    timestamp: Optional[float] = None,
                    date_str: Optional[str] = None,
                    time_str: Optional[str] = None
                    ) -> DataFolder:
        """Create a new DataFolder with a unique name within the DataStore.

        Optionally, either "timestamp" or both "date_str" and "time_str"
        may be specified to determine the date code of the folder name.
        When neither are specified, the current date and time will be used.

        Parameters:
            label: Short label describing the measurement. This label will be part of the directory name in the
                file system. It should not contain whitespace or strange characters.
            timestamp: Optional POSIX timestamp to use for the folder name.
            date_str: Optional date code to use for the folder name. If specified, it must be a string of 8 digits in
                YYYYmmdd format.
            time_str: Optional time code to use for the folder name. If specified, it must be a string of 6 digits in
                HHMMSS format.

        Returns:
            New DataFolder instance.

        Raises:
            FileExistsError: If the DataFolder already exists.
        """

        # Determine date_str and time_str.
        if (date_str is None) or (time_str is None):
            if (date_str is not None) or (time_str is not None):
                raise ValueError("Specify both date_str and time_str or neither")
            # Generate date_str and time_str from timestamp.
            if timestamp is None:
                timestamp = time.time()
            if self.USE_LOCAL_TIME:
                tm = time.localtime(timestamp)
            else:
                tm = time.gmtime(timestamp)
            date_str = time.strftime("%Y%m%d", tm)
            time_str = time.strftime("%H%M%S", tm)
        else:
            # Specified by user.
            if timestamp is not None:
                raise ValueError("Do not specify date_str, time_str and timestamp together")
            if not re.match(r"^[0-9]{8}$", date_str):
                raise ValueError("Invalid format for date_str")
            if not re.match(r"^[0-9]{6}$", time_str):
                raise ValueError("Invalid format for time_str")

        # Check label format.
        if (not isinstance(label, str)) or (not label):
            raise ValueError("Label must be a non-empty string")
        if not re.match(r"^[-_a-zA-Z0-9().,]+$", label):
            raise ValueError("Invalid characters in label")

        # Ensure date subdirectory exists.
        date_path = os.path.join(self.basedir, date_str)
        if not os.path.isdir(date_path):
            try:
                os.mkdir(date_path)
            except FileExistsError:
                # Someone created the directory just before we did ?
                # Fine, it does not matter who created it.
                pass

        # Create DataFolder directory.
        rel_path = _relative_folder_path(date_str, time_str, label)
        full_path = os.path.join(self.basedir, rel_path)
        if os.path.exists(full_path):
            raise FileExistsError("Directory {!r} already exists".format(full_path))
        os.mkdir(full_path)

        return DataFolder(full_path, label, date_str, time_str)

    def get_folder(self, label: str, date_str: str, time_str: str) -> DataFolder:
        """Open the DataFolder item with specified date code and label.

        Parameters:
            date_str: Date code of the DataFolder.
            time_str: Time code of the DataFolder.
            label: Label of the DataFolder.

        Returns:
            The matching DataFolder.
        
        Raises:
            FileNotFoundError: If the specified DataFolder does not exist.
        """

        rel_path = _relative_folder_path(date_str, time_str, label)
        full_path = os.path.join(self.basedir, rel_path)
        return DataFolder(full_path, label, date_str, time_str)

    def get_folder_from_path(self, path: str) -> DataFolder:
        """Open the DataFolder with the specified path in the filesystem.

        Parameters:
            path: Path to the DataFolder.
            The path may be either an absolute path,
            or a relative path from the DataStore base directory,
            or a relative path from the current directory.

        Returns:
            The matching DataFolder instance.

        Raises:
            FileNotFoundError: If the specified DataFolder does not exist.
        """

        if (not os.path.isabs(path)) and (not os.path.exists(path)):
            # Try to resolve the specified path relative to the base directory.
            path = os.path.join(self.basedir, path)

        if not os.path.isdir(path):
            raise FileNotFoundError("Data folder {!r} not found".format(path))

        return DataFolder(path, label=None, date_str=None, time_str=None)

    def list_folders(self, label: Optional[str] = None) -> List[DataFolder]:
        """Return a list of DataFolder items in the DataStore.

        This function may be slow when used on a large DataStore.

        Parameters:
            label: Optional folder label. When specified, only folders
            with a matching name are returned.

        Returns:
            List of matching DataFolder items.
        """

        ret = []

        date_dirs = os.listdir(self.basedir)
        date_dirs.sort()
        for dd in date_dirs:
            if re.match(r"^[0-9]{8}$", dd):
                date_str = dd
                date_path = os.path.join(self.basedir, dd)
                folders = os.listdir(date_path)
                folders.sort()
                for ff in folders:
                    m = re.match(r"^([0-9]{6})_(.+)$", ff)
                    if m:
                        time_str = m.group(1)
                        folder_label = m.group(2)
                        if (label is None) or (label == folder_label):
                            folder_path = os.path.join(date_path, ff)
                            if os.path.isdir(folder_path):
                                ret.append(DataFolder(folder_path, label, date_str, time_str))

        return ret

    def find_latest_folder(self, label: str, date_str: Optional[str] = None) -> Optional[DataFolder]:
        """Find the most recent matching DataFolder item.

        Parameters:
            label: Folder label to search for.
            date_str: Optional date code to restrict the search. When not specified, the most recent matching folder
                from any date is returned.
        Returns:
            Most recent matching DataFolder, or None if no matching DataFolder exists.
        """

        if date_str is None:
            date_dirs = os.listdir(self.basedir)
        else:
            date_dirs = [date_str]

        date_dirs.sort(reverse=True)
        for dd in date_dirs:
            if re.match(r"^[0-9]{8}$", dd):
                date_path = os.path.join(self.basedir, dd)
                folders = os.listdir(date_path)
                folders.sort(reverse=True)
                for ff in folders:
                    m = re.match(r"^([0-9]{6})_(.+)$", ff)
                    if m and m.group(2) == label:
                        time_str = m.group(1)
                        folder_path = os.path.join(date_path, ff)
                        if os.path.isdir(folder_path):
                            return DataFolder(folder_path, label, dd, time_str)

        return None



from dataclasses import asdict
import unittest
from unittest.mock import patch, ANY, create_autospec, PropertyMock

import qmi
from qmi.instruments.timebase import TimeBase_Dim3000
from qmi.instruments.timebase.dim3000 import DIM3000SweepMode, DIM3000FMDeviation
from qmi.core.exceptions import QMI_InstrumentException
from qmi.core.instrument import QMI_InstrumentIdentification
from qmi.core.transport import QMI_Transport


class TestDim3000(unittest.TestCase):

    def setUp(self) -> None:
        self.maxDiff = None

        qmi.start("TestDim3000", init_logging=False)

        # Create mocks
        self._transport_mock = create_autospec(QMI_Transport)

        # Add patches
        patcher = patch('qmi.instruments.timebase.dim3000.create_transport', return_value=self._transport_mock)
        self._transport_factory = patcher.start()
        self.addCleanup(patcher.stop)

        self.instr: TimeBase_Dim3000 = qmi.make_instrument('dim3000', TimeBase_Dim3000, "foo")
        self.instr.open()

    def tearDown(self) -> None:
        if self.instr.is_open():
            self.instr.close()
        qmi.stop()

    def test_init(self):
        expected_default_attrs = {'baudrate': 19200}
        self._transport_factory.assert_called_once_with(ANY,
                                                        default_attributes=expected_default_attrs)

    def test_open(self):
        self._transport_mock.open.assert_called_once()

    def test_close(self):
        self.instr.close()
        self._transport_mock.close.assert_called_once()

    def test_bad_string(self):
        self._transport_mock.read_until.return_value = b'error\n'
        with self.assertRaises(QMI_InstrumentException):
            self.instr.get_device_info()

    def test_get_device_info(self):
        self._transport_mock.read_until.return_value = b'Rdev:ADRV4|Rhv:4|Rfv:5|Rfb:168|Rsn:111111\n'
        dev_info = self.instr.get_device_info()
        self._transport_mock.write.assert_called_with(b'Gdev\n')
        self.assertDictEqual(
            dict(dev='ADRV4', hv='4', fv='5', fb='168', sn='111111'),
            asdict(dev_info)
        )

    def test_get_idn(self):
        expected_idn = QMI_InstrumentIdentification("TimeBase", "DIM3000[ADRV4]", "111111", "4.5.168")
        self._transport_mock.read_until.return_value = b'Rdev:ADRV4|Rhv:4|Rfv:5|Rfb:168|Rsn:111111\n'
        actual_idn = self.instr.get_idn()
        self.assertTupleEqual(expected_idn, actual_idn)

    def test_get_init_data(self):
        self._transport_mock.read_until.return_value = b'Ramoffsmin:-225|Ramoffsmax:25|Ramoffsnom:0|Rinit:1\n'
        init_data = self.instr.get_init_data()
        self._transport_mock.write.assert_called_with(b'Ginit\n')
        self.assertDictEqual(
            dict(amoffsmin=-225, amoffsmax=25, amoffsnom=0, init=True, adcoffs=None, btstat=None),
            asdict(init_data)
        )

    def test_get_init_data_btstat(self):
        self._transport_mock.read_until.return_value = b'Ramoffsmin:-225|Ramoffsmax:25|Ramoffsnom:0|Rbtstat:0|Rinit:1\n'
        init_data = self.instr.get_init_data()
        self._transport_mock.write.assert_called_with(b'Ginit\n')
        self.assertDictEqual(
            dict(amoffsmin=-225, amoffsmax=25, amoffsnom=0, btstat=False, adcoffs=None, init=True),
            asdict(init_data)
        )

    def test_get_init_data_adcoffs(self):
        self._transport_mock.read_until.return_value = b'Ramoffsmin:-225|Ramoffsmax:25|Ramoffsnom:0|Radcoffs:-24|Rinit:1\n'
        init_data = self.instr.get_init_data()
        self._transport_mock.write.assert_called_with(b'Ginit\n')
        self.assertDictEqual(
            dict(amoffsmin=-225, amoffsmax=25, amoffsnom=0, btstat=None, adcoffs=-24, init=True),
            asdict(init_data)
        )

    def test_get_parameters(self):
        self._transport_mock.read_until.return_value = (
            b'Rfreq:39195001|Rampl:236|Rout:1|Rpmon:0|Rpmfr:34|Rpmd:200|Rpmphc:0|Rswpm:0|Rswps:20000000|'
            b'Rswpp:71000222|Rswpf:700|Rswpt:45000|Rfmon:0|Rfmdev:11|Rplson:0|Rplsfr:66|Rplsdt:50|'
            b'Rffreq:32000000|Rfampl:156|Ramoffs:0|Rpcbtemp:6175|Rrefstat:0|Rreflev:-85|Rvcclev:2418\n'
        )
        init_data = self.instr.get_parameters()
        self._transport_mock.write.assert_called_with(b'Gpar\n')
        self.assertDictEqual(
            dict(freq=39195001, ampl=23.6, out=True, swpm=DIM3000SweepMode.OFF, swps=20000000,
                 swpp=71000222, swpf=700, swpt=45000, fmon=False, fmdev=DIM3000FMDeviation._6553600HZ, plson=False, plsfr=66, plsdt=50,
                 ffreq=32000000, fampl=15.6, amoffs=0, pcbtemp=61.75, refstat=False, reflev=-85, vcclev=24.18),
            asdict(init_data)
        )

    @patch.object(TimeBase_Dim3000, "MINIMUM_EXEC_DELAY_S", new_callable=PropertyMock(return_value=0.0))
    def test_setters(self, _):
        self.instr.set_output_frequency(1000)
        self._transport_mock.write.assert_called_with(b'Sfreq:1000\n')

        with self.assertRaises(ValueError):
            self.instr.set_output_frequency(TimeBase_Dim3000.FREQ_RANGE[0]-1)

        with self.assertRaises(ValueError):
            self.instr.set_output_frequency(TimeBase_Dim3000.FREQ_RANGE[1]+1)

        self.instr.set_output_amplitude(15.4)
        self._transport_mock.write.assert_called_with(b'Sampl:154\n')

        self.instr.set_sweep_mode(DIM3000SweepMode.OFF)
        self._transport_mock.write.assert_called_with(b'Sswpm:0\n')

        self.instr.set_sweep_start_frequency(1000)
        self._transport_mock.write.assert_called_with(b'Sswps:1000\n')

        with self.assertRaises(ValueError):
            self.instr.set_sweep_start_frequency(TimeBase_Dim3000.FREQ_RANGE[0]-1)

        with self.assertRaises(ValueError):
            self.instr.set_sweep_start_frequency(TimeBase_Dim3000.FREQ_RANGE[1]+1)

        self.instr.set_sweep_stop_frequency(100)
        self._transport_mock.write.assert_called_with(b'Sswpp:100\n')

        with self.assertRaises(ValueError):
            self.instr.set_sweep_stop_frequency(TimeBase_Dim3000.FREQ_RANGE[0]-1)

        with self.assertRaises(ValueError):
            self.instr.set_sweep_stop_frequency(TimeBase_Dim3000.FREQ_RANGE[1]+1)

        self.instr.set_sweep_step_frequency(100)
        self._transport_mock.write.assert_called_with(b'Sswpf:100\n')

        with self.assertRaises(ValueError):
            self.instr.set_sweep_step_frequency(TimeBase_Dim3000.FREQ_RANGE[0]-1)

        with self.assertRaises(ValueError):
            self.instr.set_sweep_step_frequency(TimeBase_Dim3000.FREQ_RANGE[1]+1)

        self.instr.set_sweep_step_time(500)
        self._transport_mock.write.assert_called_with(b'Sswpt:500\n')

        with self.assertRaises(ValueError):
            self.instr.set_sweep_step_time(TimeBase_Dim3000.TIME_RANGE[0]-1)

        with self.assertRaises(ValueError):
            self.instr.set_sweep_step_time(TimeBase_Dim3000.TIME_RANGE[1]+1)

        self.instr.enable_fm_input()
        self._transport_mock.write.assert_called_with(b'Sfmon:1\n')

        self.instr.disable_fm_input()
        self._transport_mock.write.assert_called_with(b'Sfmon:0\n')

        self.instr.set_fm_deviation(DIM3000FMDeviation._104857600HZ)
        self._transport_mock.write.assert_called_with(b'Sfmdev:15\n')

        self.instr.enable_pulse_mode()
        self._transport_mock.write.assert_called_with(b'Splson:1\n')

        self.instr.disable_pulse_mode()
        self._transport_mock.write.assert_called_with(b'Splson:0\n')

        self.instr.set_pulse_frequency(1000)
        self._transport_mock.write.assert_called_with(b'Splsfr:1000\n')

        with self.assertRaises(ValueError):
            self.instr.set_pulse_frequency(TimeBase_Dim3000.PULSE_FREQ_RANGE[0]-1)

        with self.assertRaises(ValueError):
            self.instr.set_pulse_frequency(TimeBase_Dim3000.PULSE_FREQ_RANGE[1]+1)

        self.instr.set_pulse_duty_cycle(50)
        self._transport_mock.write.assert_called_with(b'Splsdt:50\n')

        with self.assertRaises(ValueError):
            self.instr.set_pulse_duty_cycle(TimeBase_Dim3000.DUTY_CYCLE_RANGE[0]-1)

        with self.assertRaises(ValueError):
            self.instr.set_pulse_duty_cycle(TimeBase_Dim3000.DUTY_CYCLE_RANGE[1]+1)

        self.instr.set_fsk_frequency(100)
        self._transport_mock.write.assert_called_with(b'Sffreq:100\n')

        with self.assertRaises(ValueError):
            self.instr.set_fsk_frequency(TimeBase_Dim3000.FREQ_RANGE[0]-1)

        with self.assertRaises(ValueError):
            self.instr.set_fsk_frequency(TimeBase_Dim3000.FREQ_RANGE[1]+1)

        self.instr.set_fsk_amplitude(30.6)
        self._transport_mock.write.assert_called_with(b'Sfampl:306\n')

        self.instr.set_am_offset(0)
        self._transport_mock.write.assert_called_with(b'Samoffs:0\n')

        with self.assertRaises(ValueError):
            self.instr.set_am_offset(TimeBase_Dim3000.AM_OFFSET_RANGE[0]-1)

        with self.assertRaises(ValueError):
            self.instr.set_am_offset(TimeBase_Dim3000.AM_OFFSET_RANGE[1]+1)



"""
The qmi.instruments.bristol package provides support for devices manufactured by Bristol Instruments.

Currently, supported are:
- 871 Series laser wavelength meter.
- four-channel Fiber-Optic Switch (FOS)
"""

from qmi.instruments.bristol.bristol_871a import Bristol_871A
from qmi.instruments.bristol.fos import Bristol_Fos as Bristol_Fos
